# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; specifically version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# This code was inspired in the autotest project,
# client/shared/utils.py
# Authors: Plamen Dimitrov <plamen.dimitrov@intra2net.com>, Kristof Katus <kristof.katus@intra2net.com>

"""
Diff validator: Utility for testing file changes

Some typical use of this utility would be:

>>> import diff_validator
>>> change = diff_validator.Change()
>>> change.add_validated_files(["/etc/somerc"])
>>> change.append_expected_add("/etc/somerc", "this is a new line")
>>> change.append_expected_remove("/etc/somerc", "this line is removed")
>>> diff_validator.make_temp_file_copies(change.get_target_files())

After making changes through some in-test operation:

>>> changes = diff_validator.extract_changes(change.get_target_files())
>>> change_success = diff_validator.assert_change(changes, change.files_dict)

If test fails due to invalid change on the system:

>>> if not change_success:
>>>     changes = diff_validator.assert_change_dict(changes, change.files_dict)
>>>     raise DiffValidationError("Change is different than expected:\n%s" % diff_validator.create_diff_report(changes))
>>> else:
>>>     logging.info("Change made successfully")
>>> diff_validator.del_temp_file_copies(change.get_target_files())

"""

import difflib
import os
import shutil


class DiffValidationError(Exception):
    pass


def get_temp_file_path(file_path):
    """
    Generates a temporary filename.

    :param str file_path: file path prefix
    :returns: appended file path
    :rtype: str
    """
    return file_path + '.tmp'


def make_temp_file_copies(file_paths):
    """
    Creates temporary copies of the provided files.

    :param file_paths: file paths to be copied
    :type file_paths: [str]
    """
    for file_path in file_paths:
        temp_file_path = get_temp_file_path(file_path)
        shutil.copyfile(file_path, temp_file_path)


def del_temp_file_copies(file_paths):
    """
    Deletes all the provided files.

    :param file_paths: deleted file paths (their temporary versions)
    :type file_paths: [str]
    """
    for file_path in file_paths:
        temp_file_path = get_temp_file_path(file_path)
        os.remove(temp_file_path)


def parse_unified_diff_output(lines):
    """
    Parses the unified diff output of two files.

    :param lines: diff lines
    :type lines: [str]
    :returns: pair of adds and removes, where each is a list of trimmed lines
    :rtype: ([str], [str])
    """
    adds = []
    removes = []
    for line in lines:
        # ignore filepaths in the output
        if (len(line) > 2 and (line[:3] == "+++" or
                               line[:3] == "---")):
            continue
        # ignore line range information in the output
        elif len(line) > 1 and line[:2] == "@@":
            continue
        # gather adds
        elif len(line) > 0 and line[0] == "+":
            added_line = line[1:].lstrip().rstrip()
            if len(added_line) == 0:
                continue
            adds = adds + [added_line]
        # gather removes
        elif len(line) > 0 and line[0] == "-":
            removed_line = line[1:].lstrip().rstrip()
            if len(removed_line) == 0:
                continue
            removes = removes + [removed_line]
    return (adds, removes)


def extract_changes(file_paths, compared_file_paths=None):
    """
    Extracts diff information based on the new and temporarily saved old files.

    :param file_paths: original file paths (whose temporary versions will be retrieved)
    :type file_paths: [str]
    :param compared_file_paths: custom file paths to use instead of the temporary versions
    :type compared_file_paths: [str] or None
    :returns: file paths with corresponding diff information key-value pairs
    :rtype: {str, ([str], [str])}
    """
    changes = {}
    if compared_file_paths is None:
        compared_file_paths = []

    for i in range(len(file_paths)):
        temp_file_path = get_temp_file_path(file_paths[i])

        if len(compared_file_paths) > i:
            file1, file2 = compared_file_paths[i], file_paths[i]
        else:
            file1, file2 = temp_file_path, file_paths[i]
        with open(file1) as f1:
            lines1 = f1.readlines()
        with open(file2) as f2:
            lines2 = f2.readlines()
        lines = difflib.unified_diff(lines1, lines2,
                                     fromfile=file1, tofile=file2, n=0)

        changes[file_paths[i]] = parse_unified_diff_output(lines)
    return changes


def assert_change_dict(actual_result, expected_result):
    """
    Calculates unexpected line changes.

    :param actual_result: actual added and removed lines
    :type actual_result: {file_path, ([added_line, ...], [removed_line, ...])}
    :param expected_result: expected added and removed lines
    :type expected_result: {file_path, ([added_line, ...], [removed_line, ...])}
    :returns: detected differences as groups of lines with filepath keys and a tuple of
              (unexpected_adds, not_present_adds, unexpected_removes, not_present_removes)
    :rtype: {str, (str, str, str, str)}
    """
    change_diffs = {}
    for file_path, actual_changes in actual_result.items():
        expected_changes = expected_result[file_path]

        actual_adds = actual_changes[0]
        actual_removes = actual_changes[1]
        expected_adds = expected_changes[0]
        expected_removes = expected_changes[1]

        # Additional unexpected adds -- they should have been not added
        unexpected_adds = sorted(set(actual_adds) - set(expected_adds))
        # Not present expected adds -- they should have been added
        not_present_adds = sorted(set(expected_adds) - set(actual_adds))
        # Additional unexpected removes - they should have been not removed
        unexpected_removes = sorted(set(actual_removes) - set(expected_removes))
        # Not present expected removes - they should have been removed
        not_present_removes = sorted(set(expected_removes) -
                                     set(actual_removes))

        change_diffs[file_path] = (unexpected_adds, not_present_adds,
                                   unexpected_removes, not_present_removes)

    return change_diffs


def assert_change(actual_result, expected_result):
    """
    Condition wrapper of the upper method.

    :param actual_result: actual added and removed lines with filepath keys and a tuple of
                          ([added_line, ...], [removed_line, ...])
    :type actual_result: {str, ([str], [str])}
    :param expected_result: expected added and removed lines of type as the actual result
    :type expected_result: {str, ([str], [str])}
    :returns: whether changes were detected
    :rtype: bool
    """
    change_diffs = assert_change_dict(actual_result, expected_result)
    for file_change in change_diffs.values():
        for line_change in file_change:
            if len(line_change) != 0:
                return False
    return True


def create_diff_report(change_diffs):
    """
    Pretty prints the output of the `change_diffs` variable.

    :param change_diffs: detected differences as groups of lines with filepath keys and a tuple of
                         (unexpected_adds, not_present_adds, unexpected_removes, not_present_removes)
    :type: {str, (str, str, str, str)}
    :returns: print string of the line differences
    :rtype: str
    """
    diff_strings = []
    for file_path, change_diff in change_diffs.items():
        if not (change_diff[0] or change_diff[1] or
                change_diff[2] or change_diff[3]):
            continue
        diff_strings.append("--- %s" % get_temp_file_path(file_path))
        diff_strings.append("+++ %s" % file_path)
        for iter_category in range(4):
            change_category = change_diff[iter_category]
            if iter_category == 0 and change_category:
                diff_strings.append("*++ Additional unexpected adds")
            elif iter_category == 1 and change_category:
                diff_strings.append("/++ Not present expected adds")
            elif iter_category == 2 and change_category:
                diff_strings.append("*-- Additional unexpected removes")
            elif iter_category == 3 and change_category:
                diff_strings.append("/-- Not present expected removes")
            for line_change in change_category:
                diff_strings.append(str(line_change).encode('unicode_escape').decode())
    return "\n".join(diff_strings)


class Change():
    """Class for tracking and validating file changes"""

    def __init__(self):
        """Creates a change object."""
        self.files_dict = {}

    def get_target_files(self):
        """Get added files for change."""
        return list(self.files_dict.keys())

    def add_validated_files(self, filenames):
        """
        Add file to change object.

        :param filenames: files to validate
        :type filenames: [str]
        """
        for filename in filenames:
            self.files_dict[filename] = ([], [])

    def append_expected_add(self, filename, line):
        """
        Append expected added line to a file.

        :param str filename: file to append to
        :param str line: line to append to as an expected addition
        """
        try:
            self.files_dict[filename][0].append(line)
        except KeyError:
            self.files_dict[filename] = ([], [])
            self.files_dict[filename][0].append(line)

    def append_expected_remove(self, filename, line):
        """
        Append removed added line to a file.

        :param str filename: file to append to
        :param str line: line to append to as an expected removal
        """
        try:
            self.files_dict[filename][1].append(line)
        except KeyError:
            self.files_dict[filename] = ([], [])
            self.files_dict[filename][1].append(line)

    def get_all_adds(self):
        """Return a list of the added lines for all validated files."""
        all_adds = []
        for f in list(self.files_dict.keys()):
            for add in self.files_dict[f][0]:
                all_adds.append(add)
        return all_adds

    def get_all_removes(self):
        """Return a list of the removed lines for all validated files."""
        all_removes = []
        for f in list(self.files_dict.keys()):
            for add in self.files_dict[f][1]:
                all_removes.append(add)
        return all_removes



import os
import tempfile
import unittest.mock

from avocado.core import exceptions, test
from avocado.core.test_id import TestID
from avocado.utils import script

from .. import setup_avocado_loggers, temp_dir_prefix

setup_avocado_loggers()


PASS_SCRIPT_CONTENTS = """#!/bin/sh
true
"""

FAIL_SCRIPT_CONTENTS = """#!/bin/sh
false
"""


class TestClassTestUnit(unittest.TestCase):

    class DummyTest(test.Test):
        def test(self):
            pass

        def skip(self):
            self.skipTest('dummy skip test')

    def setUp(self):
        prefix = temp_dir_prefix(__name__, self, 'setUp')
        self.tmpdir = tempfile.TemporaryDirectory(prefix=prefix)

    def _get_fake_filename_test(self, name):

        class FakeFilename(test.Test):
            @property
            def filename(self):
                return name

            def test(self):
                pass

        tst_id = TestID("test", name=name)
        return FakeFilename("test", tst_id, base_logdir=self.tmpdir.name)

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_ugly_name(self):
        def run(name, path_name):
            """ Initialize test and check the dirs were created """
            tst = self.DummyTest("test", TestID(1, name),
                                 base_logdir=self.tmpdir.name)
            self.assertEqual(os.path.basename(tst.logdir), path_name)
            self.assertTrue(os.path.exists(tst.logdir))
            self.assertEqual(os.path.dirname(os.path.dirname(tst.logdir)),
                             self.tmpdir.name)

        run("/absolute/path", "1-_absolute_path")
        run("./relative/path", "1-._relative_path")
        run("../../multi_level/relative/path",
            "1-.._.._multi_level_relative_path")
        # Greek word 'kosme'
        run("\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5",
            "1-\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5")
        # Particularly problematic noncharacters in 16-bit applications
        name = ("\xb7\x95\xef\xb7\x96\xef\xb7\x97\xef\xb7\x98\xef\xb7\x99"
                "\xef\xb7\x9a\xef\xb7\x9b\xef\xb7\x9c\xef\xb7\x9d\xef\xb7"
                "\x9e\xef\xb7\x9f\xef\xb7\xa0\xef\xb7\xa1\xef\xb7\xa2\xef"
                "\xb7\xa3\xef\xb7\xa4\xef\xb7\xa5\xef\xb7\xa6\xef\xb7\xa7"
                "\xef\xb7\xa8\xef\xb7\xa9\xef\xb7\xaa\xef\xb7\xab\xef\xb7"
                "\xac\xef\xb7\xad\xef\xb7\xae\xef\xb7\xaf")
        run(name, "1-" + name)

    def test_long_name(self):
        def check(uid, name, variant, exp_logdir):
            tst = self.DummyTest("test", TestID(uid, name, variant),
                                 base_logdir=self.tmpdir.name)
            self.assertEqual(os.path.basename(tst.logdir), exp_logdir)
            return tst

        # Everything fits
        check(1, "a" * 253, None, "1-" + ("a" * 253))
        check(2, "a" * 251, {"variant_id": 1}, "2-" + ("a" * 251) + "_1")
        check(99, "a" * 249, {"variant_id": 88}, "99-" + ("a" * 249) + "_88")
        # Shrink name
        check(3, "a" * 252, {"variant_id": 1}, "3-" + ('a' * 251) + "_1")
        # Shrink variant
        check("a" * 253, "whatever", {"variant_id": 99}, "a" * 253 + "_9")
        check("a" * 254, "whatever", {"variant_id": 99}, "a" * 254 + "_")
        # No variant
        tst = check("a" * 255, "whatever", {"variant_id": "whatever-else"},
                    "a" * 255)
        # Impossible to store (uid does not fit
        self.assertRaises(RuntimeError, check, "a" * 256, "whatever",
                          {"variant_id": "else"}, None)

        self.assertEqual(os.path.basename(tst.workdir),
                         os.path.basename(tst.logdir))

    def test_data_dir(self):
        """
        Checks `get_data()` won't report fs-unfriendly data dir name
        """
        max_length_name = os.path.join(self.tmpdir.name, "a" * 250)
        tst = self._get_fake_filename_test(max_length_name)
        self.assertEqual(os.path.join(self.tmpdir.name, max_length_name + ".data"),
                         tst.get_data('', 'file', False))

    def test_no_data_dir(self):
        """
        Tests that with a filename too long, no datadir is possible
        """
        above_limit_name = os.path.join(self.tmpdir.name, "a" * 251)
        tst = self._get_fake_filename_test(above_limit_name)
        self.assertFalse(tst.get_data('', 'file', False))
        tst._record_reference('stdout', 'stdout.expected')
        tst._record_reference('stderr', 'stderr.expected')
        tst._record_reference('output', 'output.expected')

    def test_all_dirs_exists_no_hang(self):
        with unittest.mock.patch('os.path.exists', return_value=True):
            self.assertRaises(exceptions.TestSetupFail, self.DummyTest, "test",
                              TestID(1, "name"), base_logdir=self.tmpdir.name)

    def test_try_override_test_variable(self):
        dummy_test = self.DummyTest(base_logdir=self.tmpdir.name)
        self.assertRaises(AttributeError, setattr, dummy_test, "name", "whatever")
        self.assertRaises(AttributeError, setattr, dummy_test, "status", "whatever")

    def test_check_reference_success(self):
        '''
        Tests that a check is made, and is successful
        '''
        class GetDataTest(test.Test):
            def test(self):
                pass

            def get_data(self, filename, source=None, must_exist=True):
                # return the filename (path, really) unchanged
                return filename

        tst = GetDataTest("test", TestID(1, "test"),
                          base_logdir=self.tmpdir.name)
        content = 'expected content\n'
        content_path = os.path.join(tst.logdir, 'content')
        with open(content_path, 'w') as produced:
            produced.write(content)
        self.assertTrue(tst._check_reference(content_path,
                                             content_path,
                                             'content.diff',
                                             'content_diff',
                                             'Content'))

    def test_check_reference_does_not_exist(self):
        '''
        Tests that a check is not made for a file that does not exist
        '''
        tst = self.DummyTest("test", TestID(1, "test"),
                             base_logdir=self.tmpdir.name)
        self.assertFalse(tst._check_reference('does_not_exist',
                                              'stdout.expected',
                                              'stdout.diff',
                                              'stdout_diff',
                                              'Stdout'))


class TestClassTest(unittest.TestCase):

    def setUp(self):
        class AvocadoPass(test.Test):

            def test(self):
                variable = True
                self.assertTrue(variable)
                self.whiteboard = 'foo'

        prefix = temp_dir_prefix(__name__, self, 'setUp')
        self.base_logdir = tempfile.TemporaryDirectory(prefix=prefix)
        self.tst_instance_pass = AvocadoPass(base_logdir=self.base_logdir.name)
        self.tst_instance_pass.run_avocado()

    def test_class_attributes_name(self):
        self.assertEqual(self.tst_instance_pass.name, '0-AvocadoPass')

    def test_class_attributes_status(self):
        self.assertEqual(self.tst_instance_pass.status, 'PASS')

    def test_class_attributes_time_elapsed(self):
        self.assertIsInstance(self.tst_instance_pass.time_elapsed, float)

    def test_whiteboard_save(self):
        whiteboard_file = os.path.join(
            self.tst_instance_pass.logdir, 'whiteboard')
        self.assertTrue(os.path.isfile(whiteboard_file))
        with open(whiteboard_file, 'r') as whiteboard_file_obj:
            whiteboard_contents = whiteboard_file_obj.read().strip()
            self.assertTrue(whiteboard_contents, 'foo')

    def test_running_test_twice_with_the_same_uid_failure(self):
        class AvocadoPass(test.Test):

            def test(self):
                pass

        self.assertRaises(exceptions.TestSetupFail, AvocadoPass,
                          base_logdir=self.base_logdir.name)

    def tearDown(self):
        self.base_logdir.cleanup()


class SimpleTestClassTest(unittest.TestCase):

    def setUp(self):
        prefix = temp_dir_prefix(__name__, self, 'setUp')
        self.tmpdir = tempfile.TemporaryDirectory(prefix=prefix)
        self.script = None

    def test_simple_test_pass_status(self):
        self.script = script.TemporaryScript(
            'avocado_pass.sh',
            PASS_SCRIPT_CONTENTS,
            'avocado_simpletest_unittest')
        self.script.save()
        tst_instance = test.SimpleTest(
            name=TestID(1, self.script.path),
            base_logdir=self.tmpdir.name)
        tst_instance.run_avocado()
        self.assertEqual(tst_instance.status, 'PASS')

    def test_simple_test_fail_status(self):
        self.script = script.TemporaryScript(
            'avocado_fail.sh',
            FAIL_SCRIPT_CONTENTS,
            'avocado_simpletest_unittest')
        self.script.save()
        tst_instance = test.SimpleTest(
            name=TestID(1, self.script.path),
            base_logdir=self.tmpdir.name)
        tst_instance.run_avocado()
        self.assertEqual(tst_instance.status, 'FAIL')

    def tearDown(self):
        if self.script is not None:
            self.script.remove()
        self.tmpdir.cleanup()


class MockingTest(unittest.TestCase):

    def setUp(self):
        prefix = temp_dir_prefix(__name__, self, 'setUp')
        self.tmpdir = tempfile.TemporaryDirectory(prefix=prefix)

    def test_init_minimal_params(self):
        test.MockingTest(base_logdir=self.tmpdir.name)

    def test_init_positional(self):
        tst = test.MockingTest("test", TestID(1, "my_name"),
                               {}, None, "1",
                               None, None, "extra_param1",
                               "extra_param2", base_logdir=self.tmpdir.name)
        self.assertEqual(tst.name, "1-my_name")

    def test_init_kwargs(self):
        tst = test.MockingTest(methodName="test",
                               name=TestID(1, "my_name2"),
                               params={}, base_logdir=self.tmpdir.name,
                               tag="a", config=None, runner_queue=None,
                               extra1="extra_param1",
                               extra2="extra_param2")
        self.assertEqual(tst.name, "1-my_name2")

    def test_positional_kwargs(self):
        """
        Tests both positional and kwargs (theoretically impossible in
        python, but valid for nasty tests)

        keyword args are used as they explicitly represent what they mean
        """
        tst = test.MockingTest("not used", "who cares", {}, None, "0",
                               None, None, "extra_param1",
                               "extra_param2",
                               methodName="test",
                               name=TestID(1, "my_name3"),
                               params={}, base_logdir=self.tmpdir.name,
                               tag="3", config=None, runner_queue=None,
                               extra1="extra_param3",
                               extra2="extra_param4")
        self.assertEqual(tst.name, "1-my_name3")

    def test_combination(self):
        tst = test.MockingTest("test", TestID(1, "my_name4"),
                               tag="321",
                               other_param="Whatever",
                               base_logdir=self.tmpdir.name)
        self.assertEqual(tst.name, "1-my_name4")

    def test_combination_2(self):
        """
        Tests an ugly combination (positional argument overrides
        kwargs, this only happens when the substituted class reorders
        the positional arguments.  We try to first match keyword args
        and then fall-back to positional ones.
        """
        name = "positional_method_name_becomes_test_name"
        tag = "positional_base_logdir_becomes_tag"
        tst = test.MockingTest(TestID(1, name), None, None, tag,
                               methodName="test",
                               other_param="Whatever",
                               base_logdir=self.tmpdir.name)
        self.assertEqual(tst.name, "1-" + name)

    def tearDown(self):
        self.tmpdir.cleanup()


if __name__ == '__main__':
    unittest.main()



import os

from avocado.core import exit_codes
from avocado.utils import process

from .. import AVOCADO, TestCaseTmpDir

CONFIG = """[job.output.testlogs]
statuses = ["FAIL", "CANCEL"]"""


class TestLogs(TestCaseTmpDir):

    def setUp(self):
        super(TestLogs, self).setUp()
        with open(os.path.join(self.tmpdir.name, 'config'), 'w') as config:
            config.write(CONFIG)

    def test(self):
        cmd_line = ("%s --config=%s run -- examples/tests/passtest.py "
                    "examples/tests/failtest.py examples/tests/canceltest.py ")
        cmd_line %= (AVOCADO, os.path.join(self.tmpdir.name, 'config'))
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_TESTS_FAIL,
                         "Avocado did not return rc %d:\n%s"
                         % (exit_codes.AVOCADO_ALL_OK, result))
        stdout_lines = result.stdout_text.splitlines()
        self.assertNotIn('Log content for test "1-examples/tests/passtest.py'
                         ':PassTest.test" (PASS)', stdout_lines)
        self.assertIn('Log content for test "2-examples/tests/failtest.py:'
                      'FailTest.test" (FAIL)', stdout_lines)
        self.assertIn('Log content for test "3-examples/tests/canceltest.py'
                      ':CancelTest.test" (CANCEL)', stdout_lines)



#!/usr/bin/env python

import os

from avocado import Test
from avocado.utils import archive, build, process


class SyncTest(Test):

    """
    Execute the synctest test suite.

    :param sync_tarball: path to the tarball relative to a data directory
    :param default_symbols: whether to build with debug symbols (bool)
    :param sync_length: how many data should by used in sync test
    :param sync_loop: how many writes should be executed in sync test
    """

    def setUp(self):
        """
        Build the synctest suite.
        """
        self.cwd = os.getcwd()
        sync_tarball = self.params.get('sync_tarball', '*', 'synctest.tar.bz2')
        tarball_path = self.get_data(sync_tarball)
        if tarball_path is None:
            self.cancel('Test is missing data file %s' % tarball_path)
        archive.extract(tarball_path, self.workdir)
        srcdir = os.path.join(self.workdir, 'synctest')
        os.chdir(srcdir)
        if self.params.get('debug_symbols', default=True):
            build.make(srcdir,
                       env={'CFLAGS': '-g -O0'},
                       extra_args='synctest',
                       allow_output_check='none')
        else:
            build.make(srcdir,
                       allow_output_check='none')

    def test(self):
        """
        Execute synctest with the appropriate params.
        """
        path = os.path.join(os.getcwd(), 'synctest')
        cmd = ('%s %s %s' %
               (path, self.params.get('sync_length', default=100),
                self.params.get('sync_loop', default=10)))
        process.system(cmd)
        os.chdir(self.cwd)



import io
import unittest.mock

from avocado import Test
from avocado.utils import cpu


class Cpu(Test):

    @staticmethod
    def _get_file_mock(content):
        file_mock = unittest.mock.Mock()
        file_mock.__enter__ = unittest.mock.Mock(
            return_value=io.BytesIO(content))
        file_mock.__exit__ = unittest.mock.Mock()
        return file_mock

    def _get_data_mock(self, data_name):
        with open(self.get_data(data_name), 'rb') as data_file:
            content = data_file.read()
        return self._get_file_mock(content)

    def test_s390x_cpu_online(self):
        with unittest.mock.patch('avocado.utils.cpu.platform.machine',
                                 return_value='s390x'):
            with unittest.mock.patch(
                    'builtins.open',
                    return_value=self._get_data_mock('s390x')):
                self.assertEqual(len(cpu.online_list()), 2)
            with unittest.mock.patch(
                    'builtins.open',
                    return_value=self._get_data_mock('s390x_2')):
                self.assertEqual(len(cpu.online_list()), 4)

    def test_x86_64_cpu_online(self):
        with unittest.mock.patch('avocado.utils.cpu.platform.machine',
                                 return_value='x86_64'):
            with unittest.mock.patch('builtins.open',
                                     return_value=self._get_data_mock('x86_64')):
                self.assertEqual(len(cpu.online_list()), 8)

    def test_cpu_arch_i386(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('i386')):
            self.assertEqual(cpu.get_arch(), "i386")

    def test_cpu_arch_x86_64(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('x86_64')):
            self.assertEqual(cpu.get_arch(), "x86_64")

    def test_cpu_has_flags(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('x86_64')):
            self.assertTrue(cpu.cpu_has_flags('flexpriority'))
            self.assertTrue(cpu.cpu_has_flags(['sse4_2', 'xsaveopt']))
            self.assertFalse(cpu.cpu_has_flags('THIS_WILL_NEVER_BE_A_FLAG_NAME'))
            self.assertFalse(cpu.cpu_has_flags(['THIS_WILL_NEVER_BE_A_FLAG_NAME',
                                                'NEITHER_WILL_THIS_WILL_EVER_BE']))

    def test_cpu_arch_power8(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('power8')):
            self.assertEqual(cpu.get_arch(), "powerpc")

    def test_cpu_arch_power9(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('power9')):
            self.assertEqual(cpu.get_arch(), "powerpc")

    def test_cpu_arch_s390(self):
        with unittest.mock.patch(
                'builtins.open',
                return_value=self._get_data_mock('s390x')):
            self.assertEqual(cpu.get_arch(), "s390")

    def test_cpu_arch_arm_v7(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('armv7')):
            self.assertEqual(cpu.get_arch(), "arm")

    def test_cpu_arch_arm_v8(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('armv8')):
            self.assertEqual(cpu.get_arch(), "aarch64")

    def test_cpu_arch_risc_v(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('risc_v')):
            self.assertEqual(cpu.get_arch(), "riscv")

    def test_cpu_vendor_intel(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('x86_64')):
            self.assertEqual(cpu.get_vendor(), "intel")

    def test_cpu_vendor_power8(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('power8')):
            self.assertEqual(cpu.get_vendor(), "ibm")

    def test_cpu_vendor_power9(self):
        with unittest.mock.patch('builtins.open',
                                 return_value=self._get_data_mock('power9')):
            self.assertEqual(cpu.get_vendor(), "ibm")

    def test_s390x_get_version(self):
        with unittest.mock.patch('avocado.utils.cpu.platform.machine',
                                 return_value='s390x'):
            with unittest.mock.patch(
                    'builtins.open',
                    return_value=self._get_data_mock('s390x')):
                self.assertEqual(cpu.get_version(), "2827")

    def test_intel_get_version(self):
        with unittest.mock.patch('avocado.utils.cpu.platform.machine',
                                 return_value='x86_64'):
            with unittest.mock.patch('builtins.open',
                                     return_value=self._get_data_mock('x86_64')):
                self.assertEqual(cpu.get_version(), "i7-4710MQ")

    def test_power8_get_version(self):
        with unittest.mock.patch('avocado.utils.cpu.platform.machine',
                                 return_value='powerpc'):
            with unittest.mock.patch('builtins.open',
                                     return_value=self._get_data_mock('power8')):
                self.assertEqual(cpu.get_version(), "2.1")

    def test_power9_get_version(self):
        with unittest.mock.patch('avocado.utils.cpu.platform.machine',
                                 return_value='powerpc'):
            with unittest.mock.patch('builtins.open',
                                     return_value=self._get_data_mock('power9')):
                self.assertEqual(cpu.get_version(), "1.0")

    def test_s390x_get_family(self):
        with unittest.mock.patch('avocado.utils.cpu.get_arch',
                                 return_value='s390'):
            with unittest.mock.patch('avocado.utils.cpu.get_version',
                                     return_value='8561'):
                self.assertEqual(cpu.get_family(), "z15")

    def test_intel_get_family(self):
        with unittest.mock.patch('avocado.utils.cpu.get_arch',
                                 return_value='x86_64'):
            with unittest.mock.patch('avocado.utils.cpu.get_vendor',
                                     return_value='intel'):
                with unittest.mock.patch(
                        'builtins.open',
                        return_value=self._get_file_mock(b'broadwell')):
                    self.assertEqual(cpu.get_family(), "broadwell")

    def test_power8_get_family(self):
        with unittest.mock.patch('avocado.utils.cpu.get_arch',
                                 return_value='powerpc'):
            with unittest.mock.patch('builtins.open',
                                     return_value=self._get_data_mock('power8')):
                self.assertEqual(cpu.get_family(), "power8")

    def test_power9_get_family(self):
        with unittest.mock.patch('avocado.utils.cpu.get_arch', return_value='powerpc'):
            with unittest.mock.patch('builtins.open',
                                     return_value=self._get_data_mock('power9')):
                self.assertEqual(cpu.get_family(), "power9")

    def test_get_idle_state_off(self):
        retval = {0: {0: False}}
        with unittest.mock.patch('avocado.utils.cpu.online_list',
                                 return_value=[0]):
            with unittest.mock.patch('glob.glob',
                                     return_value=['/sys/devices/system/cpu/cpu0/cpuidle/state1']):
                with unittest.mock.patch('builtins.open',
                                         return_value=io.BytesIO(b'0')):
                    self.assertEqual(cpu.get_idle_state(), retval)

    def test_get_idle_state_on(self):
        retval = {0: {0: True}}
        with unittest.mock.patch('avocado.utils.cpu.online_list',
                                 return_value=[0]):
            with unittest.mock.patch('glob.glob',
                                     return_value=['/sys/devices/system/cpu/cpu0/cpuidle/state1']):
                with unittest.mock.patch('builtins.open',
                                         return_value=io.BytesIO(b'1')):
                    self.assertEqual(cpu.get_idle_state(), retval)

    def test_set_idle_state_default(self):
        output = io.BytesIO()
        with unittest.mock.patch('avocado.utils.cpu.online_list',
                                 return_value=[0]):
            with unittest.mock.patch('glob.glob',
                                     return_value=['/sys/devices/system/cpu/cpu0/cpuidle/state1']):
                with unittest.mock.patch('builtins.open',
                                         return_value=output):
                    cpu.set_idle_state()
                    self.assertEqual(output.getvalue(), b'1')

    def test_set_idle_state_withstateno(self):
        output = io.BytesIO()
        with unittest.mock.patch('avocado.utils.cpu.online_list',
                                 return_value=[0]):
            with unittest.mock.patch('glob.glob',
                                     return_value=['/sys/devices/system/cpu/cpu0/cpuidle/state2']):
                with unittest.mock.patch('builtins.open',
                                         return_value=output):
                    cpu.set_idle_state(disable=False, state_number='2')
                    self.assertEqual(output.getvalue(), b'0')

    def test_set_idle_state_withsetstate(self):
        output = io.BytesIO()
        with unittest.mock.patch('avocado.utils.cpu.online_list',
                                 return_value=[0, 2]):
            with unittest.mock.patch('glob.glob',
                                     return_value=['/sys/devices/system/cpu/cpu0/cpuidle/state1']):
                with unittest.mock.patch('builtins.open',
                                         return_value=output):
                    cpu.set_idle_state(setstate={0: {0: True}, 2: {0: False}})
                    self.assertEqual(output.getvalue(), b'10')

    def test_set_idle_state_disable(self):
        output = io.BytesIO()
        function = 'avocado.utils.cpu.online_list'
        state_file = '/sys/devices/system/cpu/cpu0/cpuidle/state1'
        with unittest.mock.patch(function, return_value=[0, 2]):
            with unittest.mock.patch('glob.glob', return_value=[state_file]):
                with unittest.mock.patch('builtins.open', return_value=output):
                    with self.assertRaises(TypeError):
                        cpu.set_idle_state(disable=1)


if __name__ == "__main__":
    unittest.main()



import os
import signal
import stat
import subprocess
import time
import unittest

import psutil

from avocado.utils import data_factory, process, script, wait

from .. import AVOCADO, BASEDIR, TestCaseTmpDir, skipOnLevelsInferiorThan

# What is commonly known as "0755" or "u=rwx,g=rx,o=rx"
DEFAULT_MODE = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
                stat.S_IRGRP | stat.S_IXGRP |
                stat.S_IROTH | stat.S_IXOTH)

BAD_TEST = """#!/usr/bin/env python
import multiprocessing
import signal
import time

def foo():
    while True:
        time.sleep(0.1)

if __name__ == "__main__":
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)
    signal.signal(signal.SIGQUIT, signal.SIG_IGN)
    proc = multiprocessing.Process(target=foo)
    proc.start()
    while True:
        time.sleep(0.1)
"""

GOOD_TEST = """#!/usr/bin/env python
import time
from avocado import Test
from avocado.core import main

class GoodTest(Test):
    def test(self):
        time.sleep(600)

if __name__ == "__main__":
    main()
"""


class InterruptTest(TestCaseTmpDir):

    @staticmethod
    def has_children(proc):
        return len(psutil.Process(proc.pid).children()) > 0

    @staticmethod
    def is_finished(proc):
        return proc.poll() is not None

    def _no_test_in_process_table(self):
        """
        Make sure the test will be really gone from the
        process table.
        """
        test_processes = []

        old_psutil = False
        try:
            process_list = psutil.pids()
        except AttributeError:
            process_list = psutil.get_pid_list()
            old_psutil = True

        for p in process_list:
            try:
                p_obj = psutil.Process(p)
                if p_obj is not None:
                    if old_psutil:
                        cmdline_list = psutil.Process(p).cmdline
                    else:
                        try:
                            cmdline_list = psutil.Process(p).cmdline()
                        except psutil.AccessDenied:
                            cmdline_list = []
                    if self.test_module in " ".join(cmdline_list):
                        test_processes.append(p_obj)
            # psutil.NoSuchProcess happens when the original
            # process already ended and left the process table
            except psutil.NoSuchProcess:
                pass

        return len(test_processes) == 0

    def setUp(self):
        super(InterruptTest, self).setUp()
        self.test_module = None

    @skipOnLevelsInferiorThan(2)
    def test_badly_behaved_sigint(self):
        """
        Make sure avocado can cleanly get out of a loop of badly behaved tests.

        :avocado: tags=parallel:1
        """
        bad_test_basename = ('wontquit-%s' %
                             data_factory.generate_random_string(5))
        bad_test = script.TemporaryScript(bad_test_basename, BAD_TEST,
                                          'avocado_interrupt_test',
                                          mode=DEFAULT_MODE)
        bad_test.save()
        self.test_module = bad_test.path
        os.chdir(BASEDIR)
        cmd = ('%s run %s --disable-sysinfo --job-results-dir %s ' %
               (AVOCADO, self.test_module, self.tmpdir.name))
        proc = subprocess.Popen(cmd.split(),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        if not wait.wait_for(lambda: self.has_children(proc), timeout=10):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado did not start the test process.')

        # This test will ignore SIGINT, so it should terminate
        # when we send the second SIGINT.
        os.kill(proc.pid, signal.SIGINT)
        # We have to actually wait 2+ seconds until
        # the ignore window is over
        time.sleep(2.5)
        os.kill(proc.pid, signal.SIGINT)

        if not wait.wait_for(lambda: self.is_finished(proc), timeout=30):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado was still running after receiving SIGINT '
                      'twice.')

        self.assertTrue(wait.wait_for(self._no_test_in_process_table,
                                      timeout=10), 'Avocado left processes behind.')

        output = proc.stdout.read()
        # Make sure the Interrupted requested sentence is there
        self.assertIn(b'Interrupt requested. Waiting 2 seconds for test to '
                      b'finish (ignoring new Ctrl+C until then)', output)
        # Make sure the Killing test subprocess message did appear
        self.assertIn(b'Killing test subprocess', output)

    @skipOnLevelsInferiorThan(2)
    def test_badly_behaved_sigterm(self):
        """
        Make sure avocado can cleanly get out of a loop of badly behaved tests.

        :avocado: tags=parallel:1
        """
        bad_test_basename = ('wontquit-%s' %
                             data_factory.generate_random_string(5))
        bad_test = script.TemporaryScript(bad_test_basename, BAD_TEST,
                                          'avocado_interrupt_test',
                                          mode=DEFAULT_MODE)
        bad_test.save()
        self.test_module = bad_test.path
        os.chdir(BASEDIR)
        cmd = ('%s run %s --disable-sysinfo --job-results-dir %s ' %
               (AVOCADO, self.test_module, self.tmpdir.name))
        proc = subprocess.Popen(cmd.split(),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        if not wait.wait_for(lambda: self.has_children(proc), timeout=10):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado did not start the test process.')

        # This test should be terminated when the main process
        # receives a SIGTERM, even if the test process ignores SIGTERM.
        os.kill(proc.pid, signal.SIGTERM)

        if not wait.wait_for(lambda: self.is_finished(proc), timeout=10):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado was still running after receiving SIGINT '
                      'twice.')

        self.assertTrue(wait.wait_for(self._no_test_in_process_table,
                                      timeout=10), 'Avocado left processes behind.')

        # Make sure the Interrupted test sentence is there
        self.assertIn(b'Terminated\n', proc.stdout.read())

    @skipOnLevelsInferiorThan(2)
    def test_well_behaved_sigint(self):
        """
        Make sure avocado can cleanly get out of a loop of well behaved tests.

        :avocado: tags=parallel:1
        """
        good_test_basename = ('goodtest-%s.py' %
                              data_factory.generate_random_string(5))
        good_test = script.TemporaryScript(good_test_basename, GOOD_TEST,
                                           'avocado_interrupt_test',
                                           mode=DEFAULT_MODE)
        good_test.save()
        self.test_module = good_test.path
        os.chdir(BASEDIR)
        cmd = ('%s run %s --disable-sysinfo --job-results-dir %s ' %
               (AVOCADO, self.test_module, self.tmpdir.name))
        proc = subprocess.Popen(cmd.split(),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        if not wait.wait_for(lambda: self.has_children(proc), timeout=10):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado did not start the test process.')

        # This test will not ignore SIGINT, so it should
        # terminate right away.
        os.kill(proc.pid, signal.SIGINT)

        if not wait.wait_for(lambda: self.is_finished(proc), timeout=10):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado was still running after receiving SIGINT '
                      'twice.')

        self.assertTrue(wait.wait_for(self._no_test_in_process_table,
                                      timeout=10), 'Avocado left processes behind.')

        output = proc.stdout.read()
        # Make sure the Interrupted requested sentence is there
        self.assertIn(b'Interrupt requested. Waiting 2 seconds for test to '
                      b'finish (ignoring new Ctrl+C until then)', output)
        # Make sure the Killing test subprocess message is not there
        self.assertNotIn(b'Killing test subprocess', output)

    @skipOnLevelsInferiorThan(2)
    def test_well_behaved_sigterm(self):
        """
        Make sure avocado can cleanly get out of a loop of well behaved tests.

        :avocado: tags=parallel:1
        """
        good_test_basename = ('goodtest-%s.py' %
                              data_factory.generate_random_string(5))
        good_test = script.TemporaryScript(good_test_basename, GOOD_TEST,
                                           'avocado_interrupt_test',
                                           mode=DEFAULT_MODE)
        good_test.save()
        self.test_module = good_test.path
        os.chdir(BASEDIR)
        cmd = ('%s run %s --disable-sysinfo --job-results-dir %s ' %
               (AVOCADO, self.test_module, self.tmpdir.name))
        proc = subprocess.Popen(cmd.split(),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        if not wait.wait_for(lambda: self.has_children(proc), timeout=10):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado did not start the test process.')

        # This test should be terminated when the main process
        # receives a SIGTERM.
        os.kill(proc.pid, signal.SIGTERM)

        if not wait.wait_for(lambda: self.is_finished(proc), timeout=10):
            process.kill_process_tree(proc.pid)
            self.fail('Avocado was still running after receiving SIGINT '
                      'twice.')

        self.assertTrue(wait.wait_for(self._no_test_in_process_table,
                                      timeout=10), 'Avocado left processes behind.')

        # Make sure the Interrupted test sentence is there
        self.assertIn(b'Terminated\n', proc.stdout.read())


if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.

"""
Maps the different status strings in avocado to booleans.

This is used by methods and functions to return a cut and dry answer to whether
a test or a job in avocado PASSed or FAILed.
"""

mapping = {"SKIP": True,
           "ABORT": False,
           "ERROR": False,
           "FAIL": False,
           "WARN": True,
           "PASS": True,
           "START": True,
           "ALERT": False,
           "RUNNING": False,
           "NOSTATUS": False,
           "INTERRUPTED": False,
           "CANCEL": True}

user_facing_status = ["SKIP",
                      "ERROR",
                      "FAIL",
                      "WARN",
                      "PASS",
                      "INTERRUPTED",
                      "CANCEL"]



import logging
import re

from ... import path as utils_path
from ... import process
from .rpm import RpmBackend

log = logging.getLogger('avocado.utils.software_manager')


class ZypperBackend(RpmBackend):

    """
    Implements the zypper backend for software manager.

    Set of operations for the zypper package manager, found on SUSE Linux.
    """

    def __init__(self):
        """
        Initializes the base command and the yum package repository.
        """
        super(ZypperBackend, self).__init__()
        self.base_command = utils_path.find_command('zypper') + ' -n'
        z_cmd = self.base_command + ' --version'
        cmd_result = process.run(z_cmd, ignore_status=True,
                                 verbose=False)
        out = cmd_result.stdout_text.strip()
        try:
            ver = re.findall(r'\d.\d*.\d*', out)[0]
        except IndexError:
            ver = out
        self.pm_version = ver
        log.debug('Zypper version: %s', self.pm_version)

    def install(self, name):
        """
        Installs package [name]. Handles local installs.

        :param name: Package Name.
        """
        i_cmd = self.base_command + ' install -l ' + name
        try:
            process.system(i_cmd, sudo=True)
            return True
        except process.CmdError:
            return False

    def add_repo(self, url):
        """
        Adds repository [url].

        :param url: URL for the package repository.
        """
        ar_cmd = self.base_command + ' addrepo ' + url
        try:
            process.system(ar_cmd, sudo=True)
            return True
        except process.CmdError:
            return False

    def remove_repo(self, url):
        """
        Removes repository [url].

        :param url: URL for the package repository.
        """
        rr_cmd = self.base_command + ' removerepo ' + url
        try:
            process.system(rr_cmd, sudo=True)
            return True
        except process.CmdError:
            return False

    def remove(self, name):
        """
        Removes package [name].
        """
        r_cmd = self.base_command + ' ' + 'erase' + ' ' + name

        try:
            process.system(r_cmd, sudo=True)
            return True
        except process.CmdError:
            return False

    def upgrade(self, name=None):
        """
        Upgrades all packages of the system.

        Optionally, upgrade individual packages.

        :param name: Optional parameter wildcard spec to upgrade
        :type name: str
        """
        if not name:
            u_cmd = self.base_command + ' update -l'
        else:
            u_cmd = self.base_command + ' ' + 'update' + ' ' + name

        try:
            process.system(u_cmd, sudo=True)
            return True
        except process.CmdError:
            return False

    def provides(self, name):
        """
        Searches for what provides a given file.

        :param name: File path.
        """
        p_cmd = self.base_command + ' what-provides ' + name
        list_provides = []
        try:
            p_output = process.system_output(p_cmd).split('\n')[4:]
            for line in p_output:
                line = [a.strip() for a in line.split('|')]
                try:
                    # state, pname, type, version, arch, repository = line
                    pname = line[1]
                    if pname not in list_provides:
                        list_provides.append(pname)
                except IndexError:
                    pass
            if len(list_provides) > 1:
                log.warning('More than one package found, '
                            'opting by the first queue result')
            if list_provides:
                log.info("Package %s provides %s", list_provides[0], name)
                return list_provides[0]
            return None
        except process.CmdError:
            return None

    def build_dep(self, name):
        """Return True if build-dependencies are installed for provided package

        Keyword argument:
        name -- name of the package
        """
        s_cmd = '%s source-install -d %s' % (self.base_command, name)

        try:
            process.system(s_cmd, sudo=True)
            return True
        except process.CmdError:
            log.error('Installing dependencies failed')
            return False

    def _source_install(self, name):
        """
        Source install the given package [name]
        Returns the SPEC file of the package

        :param name: name of the package

        :return path: path of the spec file
        """
        s_cmd = '%s source-install %s' % (self.base_command, name)

        try:
            process.system(s_cmd, sudo=True)
            if self.build_dep(name):
                return '/usr/src/packages/SPECS/%s.spec' % name
        except process.CmdError:
            log.error('Installing source failed')
            return ""

    def get_source(self, name, dest_path):
        """
        Downloads the source package and prepares it in the given dest_path
        to be ready to build

        :param name: name of the package
        :param dest_path: destination_path

        :return final_dir: path of ready-to-build directory
        """
        if not self.check_installed("rpm-build"):
            if not self.install("rpm-build"):
                log.error("SoftwareManager (RpmBackend) can't get packages"
                          "with dependency resolution: Package 'rpm-build'"
                          "could not be installed")
                return ""
        try:
            spec_path = self._source_install(name)
            if spec_path:
                return self.prepare_source(spec_path, dest_path)
            else:
                log.error("Source not installed properly")
                return ""
        except process.CmdError as details:
            log.error(details)
            return ""



from avocado.core.output import LOG_UI
from avocado.core.plugin_interfaces import CLICmd


class HelloWorld(CLICmd):

    name = 'hello'
    description = 'The classical Hello World! plugin example.'

    def run(self, config):
        LOG_UI.info(self.description)



import os
import unittest

from avocado.core import exit_codes
from avocado.utils import process

from .. import AVOCADO, BASEDIR, TestCaseTmpDir


class GetData(TestCaseTmpDir):

    def test(self):
        test_path = os.path.join(BASEDIR, "selftests", ".data", "get_data.py")
        cmd_line = "%s run --disable-sysinfo --job-results-dir '%s' -- %s"
        cmd_line %= (AVOCADO, self.tmpdir.name, test_path)
        result = process.run(cmd_line)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_ALL_OK)

    def test_varianter(self):
        test_path = os.path.join(BASEDIR, "selftests", ".data", "get_data.py")
        test_variants_path = os.path.join(BASEDIR, "selftests", ".data",
                                          "get_data.py.data", "get_data.json")
        cmd_line = ("%s run --disable-sysinfo --job-results-dir '%s' "
                    "--json-variants-load %s -- %s")
        cmd_line %= (AVOCADO, self.tmpdir.name, test_variants_path, test_path)
        result = process.run(cmd_line)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_ALL_OK)


if __name__ == '__main__':
    unittest.main()



import unittest.mock

from avocado.utils import memory


class Test(unittest.TestCase):

    def test_numa_nodes_with_memory(self):
        file_values = [u"0\n", u"1-3", u"0-1,12-14\n"]
        expected_values = [[0], [1, 2, 3], [0, 1, 12, 13, 14]]
        for value, exp in zip(file_values, expected_values):
            with unittest.mock.patch('os.path.exists', return_value=True):
                with unittest.mock.patch('avocado.utils.genio.read_file',
                                         return_value=value):
                    self.assertEqual(memory.numa_nodes_with_memory(), exp)


BUDDY_INFO_RESPONSE = (
    """Node 0, zone      DMA      1      1      0      0      1      1
Node 0, zone    DMA32    987    679   1004   3068   2795   1432
Node 1, zone   Normal   5430   9759   9044   9751  16482   8924""")


@unittest.mock.patch('avocado.utils.memory._get_buddy_info_content',
                     return_value=BUDDY_INFO_RESPONSE)
class GetBuddyInfo(unittest.TestCase):

    def test_simple_chunk_size(self, buddy_mocked):
        chunk_size = '0'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result[chunk_size], 6418)
        self.assertTrue(buddy_mocked.called)

    def test_less_than_chunk_size(self, buddy_mocked):
        chunk_size = '<2'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result['0'], 6418)
        self.assertEqual(result['1'], 10439)
        self.assertTrue(buddy_mocked.called)

    def test_less_than_equal_chunk_size(self, buddy_mocked):
        chunk_size = '<=2'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result['0'], 6418)
        self.assertEqual(result['1'], 10439)
        self.assertEqual(result['2'], 10048)
        self.assertTrue(buddy_mocked.called)

    def test_greater_than_chunk_size(self, buddy_mocked):
        chunk_size = '>3'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result['4'], 19278)
        self.assertEqual(result['5'], 10357)
        self.assertTrue(buddy_mocked.called)

    def test_greater_than_equal_chunk_size(self, buddy_mocked):
        chunk_size = '>=3'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result['3'], 12819)
        self.assertEqual(result['4'], 19278)
        self.assertEqual(result['5'], 10357)
        self.assertTrue(buddy_mocked.called)

    def test_multiple_chunk_size(self, buddy_mocked):
        chunk_size = '2 4'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result['2'], 10048)
        self.assertEqual(result['4'], 19278)
        self.assertTrue(buddy_mocked.called)

    def test_multiple_chunk_size_filtering_simple(self, buddy_mocked):
        chunk_size = '>2 <4'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result['3'], 12819)
        self.assertTrue(buddy_mocked.called)

    def test_multiple_chunk_size_filtering(self, buddy_mocked):
        chunk_size = '>=2 <=4'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result['2'], 10048)
        self.assertEqual(result['3'], 12819)
        self.assertEqual(result['4'], 19278)
        self.assertTrue(buddy_mocked.called)

    def test_multiple_chunk_size_filtering_invalid(self, buddy_mocked):
        chunk_size = '>2 <2'
        result = memory.get_buddy_info(chunk_size)
        self.assertEqual(result, {})
        self.assertTrue(buddy_mocked.called)

    def test_filtering_node(self, buddy_mocked):
        chunk_size = '0'
        result = memory.get_buddy_info(chunk_size, nodes='1')
        self.assertEqual(result[chunk_size], 5430)
        self.assertTrue(buddy_mocked.called)

    def test_filtering_zone(self, buddy_mocked):
        chunk_size = '0'
        result = memory.get_buddy_info(chunk_size, zones='DMA32')
        self.assertEqual(result[chunk_size], 987)
        self.assertTrue(buddy_mocked.called)


if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat, Inc. 2020
# Authors: Cleber Rosa <crosa@redhat.com>

"""Replay Job Plugin"""

import json
import os
import sys

from avocado.core import exit_codes, job, output
from avocado.core.data_dir import get_job_results_dir
from avocado.core.plugin_interfaces import CLICmd
from avocado.core.settings import settings


class Replay(CLICmd):

    """Implements the avocado 'replay' subcommand."""

    name = 'replay'
    description = "Runs a new job using a previous job as its configuration"

    def configure(self, parser):
        parser = super(Replay, self).configure(parser)
        help_msg = ('Replays a job, identified by: complete or partial Job '
                    'ID, "latest" for the latest job, the job results path.')
        settings.register_option(section='job.replay',
                                 key='source_job_id',
                                 help_msg=help_msg,
                                 metavar='SOURCE_JOB_ID',
                                 default='latest',
                                 nargs='?',
                                 positional_arg=True,
                                 parser=parser)

    @staticmethod
    def _exit_fail(message):
        output.LOG_UI.error(message)
        sys.exit(exit_codes.AVOCADO_FAIL)

    @staticmethod
    def _retrieve_source_job_config(source_job_id):
        results_dir = get_job_results_dir(source_job_id)
        if not results_dir:
            msg = 'Could not find the results directory for Job "%s"' % source_job_id
            Replay._exit_fail(msg)
        config_file_path = os.path.join(results_dir, 'jobdata', 'args.json')
        try:
            with open(config_file_path, 'r') as config_file:
                return json.load(config_file)
        except OSError:
            msg = 'Could not open the source Job configuration "%s"' % config_file_path
            Replay._exit_fail(msg)
        except json.decoder.JSONDecodeError:
            msg = 'Could not read a valid configuration from file "%s"' % config_file_path
            Replay._exit_fail(msg)

    def run(self, config):
        namespace = 'job.replay.source_job_id'
        source_job_id = config.get(namespace)
        source_job_config = self._retrieve_source_job_config(source_job_id)
        if hasattr(source_job_config, namespace):
            del(source_job_config[namespace])
        # Flag that this is indeed a replayed job, which is impossible to
        # tell solely based on the job.replay.source_job_id given that it
        # has a default value of 'latest' for convenience reasons
        source_job_config['job.replay.enabled'] = True
        with job.Job.from_config(source_job_config) as job_instance:
            job_run = job_instance.run()
        return job_run



import smtplib
from email.mime.text import MIMEText

from avocado.core.output import LOG_UI
from avocado.core.plugin_interfaces import Init, JobPost, JobPre
from avocado.core.settings import settings


class MailInit(Init):
    name = 'mail-init'
    description = 'Mail plugin initialization'

    def initialize(self):
        help_msg = 'Mail recipient.'
        settings.register_option(section='plugins.job.mail',
                                 key='recipient',
                                 default='root@localhost.localdomain',
                                 help_msg=help_msg)

        help_msg = 'Mail header.'
        settings.register_option(section='plugins.job.mail',
                                 key='header',
                                 default='[AVOCADO JOB NOTIFICATION]',
                                 help_msg=help_msg)

        help_msg = 'Mail sender.'
        settings.register_option(section='plugins.job.mail',
                                 key='sender',
                                 default='avocado@localhost.localdomain',
                                 help_msg=help_msg)

        help_msg = 'Mail server.'
        settings.register_option(section='plugins.job.mail',
                                 key='server',
                                 default='localhost',
                                 help_msg=help_msg)


class Mail(JobPre, JobPost):
    name = 'mail'
    description = 'Sends mail to notify on job start/end'

    def mail(self, job):
        rcpt = job.config.get('plugins.job.mail.recipient')
        header = job.config.get('plugins.job.mail.header')
        sender = job.config.get('plugins.job.mail.sender')
        server = job.config.get('plugins.job.mail.server')
        # build proper subject based on job status
        subject = '%s Job %s - Status: %s' % (header,
                                              job.unique_id,
                                              job.status)
        msg = MIMEText(subject)
        msg['Subject'] = subject
        msg['From'] = sender
        msg['To'] = rcpt

        # So many possible failures, let's just tell the user about it
        try:
            smtp = smtplib.SMTP(server)
            smtp.sendmail(sender, [rcpt], msg.as_string())
            smtp.quit()
        except Exception:  # pylint: disable=W0703
            LOG_UI.error("Failure to send email notification: "
                         "please check your mail configuration")

    pre = post = mail



from unittest import TestCase

from avocado.core.status import repo, utils


class StatusRepo(TestCase):

    def setUp(self):
        self.status_repo = repo.StatusRepo()

    def test_process_raw_message_invalid(self):
        with self.assertRaises(utils.StatusMsgInvalidJSONError):
            self.status_repo.process_raw_message('+-+-InvalidJSON-AFAICT-+-+')

    def test_process_raw_message_no_id(self):
        msg = ('{"status": "finished", "time": 1597774000.5140226, '
               '"returncode": 0}')
        with self.assertRaises(repo.StatusMsgMissingDataError):
            self.status_repo.process_raw_message(msg)

    def test_set_task_data(self):
        self.status_repo._set_task_data({"id": "1-foo", "status": "started"})
        self.assertEqual(self.status_repo._all_data["1-foo"],
                         [{"status": "started"}])

    def test_handle_task_started(self):
        msg = {"id": "1-foo", "status": "started", "output_dir": "/fake/path"}
        self.status_repo._handle_task_started(msg)
        self.assertEqual(self.status_repo.get_task_data("1-foo"),
                         [{"status": "started", "output_dir": "/fake/path"}])

    def test_handle_task_started_no_output_dir(self):
        msg = {"id": "1-foo", "status": "started"}
        with self.assertRaises(repo.StatusMsgMissingDataError):
            self.status_repo._handle_task_started(msg)

    def test_handle_task_finished_no_result(self):
        msg = {"id": "1-foo", "status": "finished"}
        self.status_repo._handle_task_finished(msg)
        self.assertEqual(self.status_repo.get_task_data("1-foo"),
                         [{"status": "finished"}])
        self.assertEqual(self.status_repo._by_result.get(None), ["1-foo"])

    def test_handle_task_finished_result(self):
        msg = {"id": "1-foo", "status": "finished", "result": "pass"}
        self.status_repo._handle_task_finished(msg)
        self.assertEqual(self.status_repo.get_task_data("1-foo"),
                         [{"status": "finished", "result": "pass"}])
        self.assertEqual(self.status_repo._by_result.get("pass"), ["1-foo"])

    def test_process_message_running(self):
        msg = {"id": "1-foo", "status": "running"}
        self.status_repo.process_message(msg)
        self.assertEqual(self.status_repo.get_task_data("1-foo"),
                         [{"status": "running"}])

    def test_process_raw_message_task_started(self):
        msg = '{"id": "1-foo", "status": "started", "output_dir": "/fake/path"}'
        self.status_repo.process_raw_message(msg)
        self.assertEqual(self.status_repo.get_task_data("1-foo"),
                         [{"status": "started", "output_dir": "/fake/path"}])

    def test_process_raw_message_task_running(self):
        msg = '{"id": "1-foo", "status": "running"}'
        self.status_repo.process_raw_message(msg)
        self.assertEqual(self.status_repo.get_task_data("1-foo"),
                         [{"status": "running"}])

    def test_process_messages_running(self):
        msg = {"id": "1-foo", "status": "running", "time": 1597894378.6080744}
        self.status_repo.process_message(msg)
        msg = {"id": "1-foo", "status": "running", "time": 1597894378.6103745}
        self.status_repo.process_message(msg)
        self.assertEqual(self.status_repo.get_latest_task_data("1-foo"),
                         {"status": "running", "time": 1597894378.6103745})

    def test_task_status_time(self):
        msg = {"id": "1-foo", "status": "running", "time": 1597894378.0000002}
        self.status_repo.process_message(msg)
        self.assertEqual(self.status_repo._status["1-foo"],
                         ("running", 1597894378.0000002))
        msg = {"id": "1-foo", "status": "started", "time": 1597894378.0000001,
               "output_dir": "/fake/path"}
        self.status_repo.process_message(msg)
        self.assertEqual(self.status_repo._status["1-foo"],
                         ("running", 1597894378.0000002))

    def test_no_task_status(self):
        self.assertIsNone(self.status_repo.get_task_status("1-no-existing-id"))

    def test_get_task_status(self):
        msg = {"id": "1-foo", "status": "running", "time": 1597894378.0000002}
        self.status_repo.process_message(msg)
        self.assertEqual(self.status_repo.get_task_status("1-foo"), "running")
        msg = {"id": "1-foo", "status": "started", "time": 1597894378.0000001,
               "output_dir": "/fake/path"}
        self.status_repo.process_message(msg)
        self.assertEqual(self.status_repo.get_task_status("1-foo"), "running")

    def test_get_task_status_journal_summary(self):
        msg = {"id": "1-foo", "status": "running", "time": 1000000003.0}
        self.status_repo.process_message(msg)
        msg = {"id": "1-foo", "status": "running", "time": 1000000002.0}
        self.status_repo.process_message(msg)
        msg = {"id": "1-foo", "status": "started", "time": 1000000001.0,
               "output_dir": "/fake/path"}
        self.status_repo.process_message(msg)
        msg = {"id": "1-foo", "status": "finished", "time": 1000000004.0,
               "result": "pass"}
        self.status_repo.process_message(msg)
        self.assertEqual(self.status_repo.get_task_status("1-foo"), "finished")
        self.assertEqual(self.status_repo.status_journal_summary.pop(),
                         ("1-foo", "finished", 1000000004.0))
        self.assertEqual(self.status_repo.status_journal_summary.pop(),
                         ("1-foo", "running", 1000000003.0))
        with self.assertRaises(IndexError):
            self.status_repo.status_journal_summary.pop()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2016
# Author: Amador Pahim <apahim@redhat.com>

import argparse
import json
import os
import re
import sys

from avocado.core import data_dir, exit_codes, jobdata, teststatus
from avocado.core.output import LOG_UI
from avocado.core.plugin_interfaces import CLI
from avocado.core.settings import settings
from avocado.core.test import ReplaySkipTest


class Replay(CLI):

    """
    Replay a job
    """

    name = 'replay'
    description = "Replay options for 'run' subcommand"

    def configure(self, parser):
        run_subcommand_parser = parser.subcommands.choices.get('run', None)
        if run_subcommand_parser is None:
            return

        msg = 'job replay'
        replay_parser = run_subcommand_parser.add_argument_group(msg)

        help_msg = ('Replay a job identified by its (partial) hash id. '
                    'Use "--replay" latest to replay the latest job.')
        settings.register_option(section='run.replay',
                                 key='job_id',
                                 default=None,
                                 help_msg=help_msg,
                                 parser=replay_parser,
                                 long_arg='--replay')

        help_msg = 'Filter tests to replay by test status.'
        settings.register_option(section='run.replay',
                                 key='test_status',
                                 default=[],
                                 help_msg=help_msg,
                                 key_type=self._valid_status,
                                 parser=replay_parser,
                                 long_arg='--replay-test-status')

        help_msg = 'Ignore variants and/or configuration from the source job.'
        settings.register_option(section='run.replay',
                                 key='ignore',
                                 default=[],
                                 help_msg=help_msg,
                                 key_type=self._valid_ignore,
                                 parser=replay_parser,
                                 long_arg='--replay-ignore')

        help_msg = 'Resume an interrupted job'
        settings.register_option(section='run.replay',
                                 key='resume',
                                 default=False,
                                 help_msg=help_msg,
                                 key_type=bool,
                                 parser=replay_parser,
                                 long_arg='--replay-resume')

    def _valid_status(self, string):
        if not string:
            return []
        status_list = string.split(',')
        for item in status_list:
            if item not in teststatus.user_facing_status:
                msg = ('Invalid --replay-test-status option. Valid '
                       'options are (more than one allowed): %s' %
                       ','.join([item for item
                                 in teststatus.user_facing_status]))
                raise argparse.ArgumentTypeError(msg)

        return status_list

    def _valid_ignore(self, string):
        if not string:
            return []
        options = ['variants', 'config']
        ignore_list = string.split(',')
        for item in ignore_list:
            if item not in options:
                msg = ('Invalid --replay-ignore option. Valid '
                       'options are (more than one allowed): %s'
                       % ','.join(options))
                raise argparse.ArgumentTypeError(msg)

        return ignore_list

    def load_config(self, resultsdir):
        config = jobdata.retrieve_config(resultsdir)
        if config is not None:
            settings.process_config_path(config)

    def _get_tests_from_tap(self, path):
        if not os.path.exists(path):
            return None
        re_result = re.compile(r"(not )?ok (\d+) ([^#]*)(# (\w+).*)?")
        re_no_tests = re.compile(r"1..(\d+)")
        max_index = 0
        no_tests = 0
        _tests = {}
        with open(path) as tapfile:
            for line in tapfile:
                line = line.strip()
                if line.startswith("#"):
                    continue
                result = re_result.match(line)
                if result:
                    if result.group(1) is None:
                        res = result.group(5)
                        if res is None:
                            res = "PASS"
                    else:
                        res = "ERROR"
                    index = int(result.group(2))
                    _tests[index] = {"status": res,
                                     "test": result.group(3).rstrip()}
                    max_index = max(max_index, index)
                    continue
                _no_tests = re_no_tests.match(line)
                if _no_tests:
                    no_tests = int(_no_tests.group(1))
                    continue

        if not (no_tests or max_index):
            return None

        # Now add _tests that were not executed
        skipped_test = {"test": "UNKNOWN", "status": "INTERRUPTED"}
        return [_tests[i] if i in _tests else skipped_test
                for i in range(1, max(max_index, no_tests) + 1)]

    def _create_replay_map(self, resultsdir, replay_filter):
        """
        Creates a mapping to be used as filter for the replay. Given
        the replay_filter, tests that should be filtered out will have a
        correspondent ReplaySkipTest class in the map. Tests that should
        be replayed will have a correspondent None in the map.
        """
        json_results = os.path.join(resultsdir, "results.json")
        if os.path.exists(json_results):
            with open(json_results, 'r') as json_file:
                results = json.loads(json_file.read())
                tests = results["tests"]
                for _ in range(results["total"] - len(tests)):
                    tests.append({"test": "UNKNOWN", "status": "INTERRUPTED"})
        else:
            # get partial results from tap
            tests = self._get_tests_from_tap(os.path.join(resultsdir,
                                                          "results.tap"))
            if not tests:   # tests not available, ignore replay map
                return None

        replay_map = []
        for test in tests:
            if test['status'] not in replay_filter:
                replay_map.append(ReplaySkipTest)
            else:
                replay_map.append(None)

        return replay_map

    def run(self, config):
        job_id = config.get('run.replay.job_id')
        if job_id is None:
            return

        err = None
        replay_ignore = config.get('run.replay.ignore')
        test_status = config.get('run.replay.test_status')

        if test_status and 'variants' in replay_ignore:
            err = ("Option `--replay-test-status` is incompatible with "
                   "`--replay-ignore variants`.")
        elif test_status and config.get('run.references'):
            err = ("Option --replay-test-status is incompatible with "
                   "test references given on the command line.")
        elif config.get("remote_hostname", False):
            err = "Currently we don't replay jobs in remote hosts."
        if err is not None:
            LOG_UI.error(err)
            sys.exit(exit_codes.AVOCADO_FAIL)

        resultsdir = data_dir.get_job_results_dir(job_id,
                                                  config.get('run.results_dir'))
        if resultsdir is None:
            LOG_UI.error("Can't find job results directory for '%s'", job_id)
            sys.exit(exit_codes.AVOCADO_FAIL)

        with open(os.path.join(resultsdir, 'id'), 'r') as id_file:
            config['replay_sourcejob'] = id_file.read().strip()

        replay_config = jobdata.retrieve_job_config(resultsdir)
        overridables = ['loaders',
                        'external_runner',
                        'external_runner_testdir',
                        'external_runner_chdir',
                        'failfast',
                        'ignore_missing_references',
                        'execution_order']
        if replay_config is None:
            LOG_UI.warn('Source job config data not found. These options will '
                        'not be loaded in this replay job: %s',
                        ', '.join(overridables))
        else:
            for option in overridables:
                optvalue = config.get(option, None)
                # Temporary, this will be removed soon
                if option in ['failfast',
                              'ignore_missing_references',
                              'execution_order',
                              'loaders',
                              'external_runner',
                              'external_runner_chdir',
                              'external_runner_testdir']:
                    optvalue = config.get('run.{}'.format(option))
                if optvalue is not None:
                    LOG_UI.warn("Overriding the replay %s with the --%s value "
                                "given on the command line.",
                                option.replace('_', '-'),
                                option.replace('_', '-'))
                elif option in replay_config:
                    config[option] = replay_config[option]

        if config.get('run.references'):
            LOG_UI.warn('Overriding the replay test references with test '
                        'references given in the command line.')
        else:
            references = jobdata.retrieve_references(resultsdir)
            if references is None:
                LOG_UI.error('Source job test references data not found. '
                             'Aborting.')
                sys.exit(exit_codes.AVOCADO_FAIL)
            else:
                config['run.references'] = references

        if 'config' in replay_ignore:
            LOG_UI.warn("Ignoring configuration from source job with "
                        "--replay-ignore.")
        else:
            self.load_config(resultsdir)

        if 'variants' in replay_ignore:
            LOG_UI.warn("Ignoring variants from source job with "
                        "--replay-ignore.")
        else:
            variants = jobdata.get_variants_path(resultsdir)
            if variants is None:
                LOG_UI.error('Source job variants data not found. Aborting.')
                sys.exit(exit_codes.AVOCADO_FAIL)
            else:
                LOG_UI.warning("Using src job variants data only, use "
                               "`--replay-ignore variants` to override "
                               "them.")
                config["json.variants.load"] = variants

        # Extend "replay_test_status" of "INTERRUPTED" when --replay-resume
        # supplied.
        if config.get('run.replay.resume'):
            if not test_status:
                config['replay_teststatus'] = ["INTERRUPTED"]
            elif "INTERRUPTED" not in test_status:
                config['replay_teststatus'].append("INTERRUPTED")
        if test_status:
            replay_map = self._create_replay_map(resultsdir,
                                                 test_status)
            config['replay_map'] = replay_map

        # Use the original directory to resolve test references properly
        pwd = jobdata.retrieve_pwd(resultsdir)
        if pwd is not None:
            if os.path.exists(pwd):
                os.chdir(pwd)
            else:
                LOG_UI.warn("Directory used in the replay source job '%s' does"
                            " not exist, using '.' instead", pwd)



import os
import shutil
import sys

from avocado import Test
from avocado.utils import build, process


class DoubleFreeTest(Test):

    """
    Double free test case.

    :avocado: tags=requires_c_compiler

    :param source: name of the source file located in a data directory
    """

    def setUp(self):
        """
        Build 'doublefree'.
        """
        source = self.params.get('source', default='doublefree.c')
        c_file = self.get_data(source)
        if c_file is None:
            self.cancel('Test is missing data file %s' % source)
        c_file_name = os.path.basename(c_file)
        dest_c_file = os.path.join(self.workdir, c_file_name)
        shutil.copy(c_file, dest_c_file)
        build.make(self.workdir,
                   env={'CFLAGS': '-g -O0'},
                   extra_args='doublefree')

    def test(self):
        """
        Execute 'doublefree'.
        """
        cmd = os.path.join(self.workdir, 'doublefree')
        cmd_result = process.run(cmd, ignore_status=True,
                                 env={'MALLOC_CHECK_': '1'})
        self.log.info(cmd_result)
        output = cmd_result.stdout + cmd_result.stderr
        if sys.platform.startswith('darwin'):
            pattern = b'pointer being freed was not allocated'
        else:
            pattern = b'free(): invalid pointer'
        self.assertTrue(pattern in output,
                        msg='Could not find pattern %s in output %s' %
                            (pattern, output))



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Exception classes, useful for tests, and other parts of the framework code.
"""


class JobBaseException(Exception):

    """
    The parent of all job exceptions.

    You should be never raising this, but just in case, we'll set its
    status' as FAIL.
    """
    status = "FAIL"


class JobError(JobBaseException):

    """
    A generic error happened during a job execution.
    """
    status = "ERROR"


class JobTestSuiteError(JobBaseException):

    """
    Generic error happened during the creation of a job's test suite
    """
    status = "ERROR"


class JobTestSuiteEmptyError(JobTestSuiteError):

    """
    Error raised when the creation of a test suite results in an empty suite
    """
    status = "ERROR"


class JobTestSuiteReferenceResolutionError(JobTestSuiteError):

    """
    Test References did not produce a valid reference by any resolver
    """
    status = "ERROR"


class OptionValidationError(Exception):

    """
    An invalid option was passed to the test runner
    """
    status = "ERROR"


class TestBaseException(Exception):

    """
    The parent of all test exceptions.

    You should be never raising this, but just in case, we'll set its
    status' as FAIL.
    """
    status = "FAIL"


class TestSetupFail(TestBaseException):

    """
    Indicates an error during a setup or cleanup procedure.
    """
    status = "ERROR"


class TestError(TestBaseException):

    """
    Indicates that the test was not fully executed and an error happened.

    This is the sort of exception you raise if the test was partially
    executed and could not complete due to a setup, configuration,
    or another fatal condition.
    """
    status = "ERROR"


class TestNotFoundError(TestBaseException):

    """
    Indicates that the test was not found in the test directory.
    """
    status = "ERROR"


class TestTimeoutInterrupted(TestBaseException):

    """
    Indicates that the test did not finish before the timeout specified.
    """
    status = "INTERRUPTED"


class TestInterruptedError(TestBaseException):

    """
    Indicates that the test was interrupted by the user (Ctrl+C)
    """
    status = "INTERRUPTED"


class TestAbortError(TestBaseException):

    """
    Indicates that the test was prematurely aborted.
    """
    status = "ERROR"


class TestSkipError(TestBaseException):

    """
    Indicates that the test is skipped.

    Should be thrown when various conditions are such that the test is
    inappropriate. For example, inappropriate architecture, wrong OS version,
    program being tested does not have the expected capability (older version).
    """
    status = "SKIP"


class TestFail(TestBaseException, AssertionError):

    """
    Indicates that the test failed.

    TestFail inherits from AssertionError in order to keep compatibility
    with vanilla python unittests (they only consider failures the ones
    deriving from AssertionError).
    """
    status = "FAIL"


class TestWarn(TestBaseException):

    """
    Indicates that bad things (may) have happened, but not an explicit
    failure.
    """
    status = "WARN"


class TestCancel(TestBaseException):
    """
    Indicates that a test was canceled.

    Should be thrown when the cancel() test method is used.
    """
    status = "CANCEL"



import json
import sys

from avocado import Test, fail_on
from avocado.utils import process


class Interface(Test):

    def get_runner(self):
        default_runner = "%s -m avocado.core.nrunner" % sys.executable
        return self.params.get("runner", default=default_runner)

    @fail_on(process.CmdError)
    def test_help(self):
        """
        Makes sure a runner can be called with --help and that the
        basic required commands are present in the help message
        """
        cmd = "%s --help" % self.get_runner()
        result = process.run(cmd)
        self.assertIn(b"capabilities", result.stdout,
                      "Mention to capabilities command not found")

    @fail_on(process.CmdError)
    def test_capabilities(self):
        cmd = "%s capabilities" % self.get_runner()
        result = process.run(cmd)
        capabilities = json.loads(result.stdout_text)
        self.assertIn("runnables", capabilities)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Manages output and logging in avocado applications.
"""
import errno
import logging
import os
import re
import sys
import traceback

from ..utils import path as utils_path
from . import exit_codes
from .settings import settings
from .streams import BUILTIN_STREAMS

#: Handle cases of logging exceptions which will lead to recursion error
logging.raiseExceptions = False

#: Pre-defined Avocado human UI logger
LOG_UI = logging.getLogger("avocado.app")
#: Pre-defined Avocado job/test logger
LOG_JOB = logging.getLogger("avocado.test")


class TermSupport:

    COLOR_BLUE = '\033[94m'
    COLOR_GREEN = '\033[92m'
    COLOR_YELLOW = '\033[93m'
    COLOR_RED = '\033[91m'
    COLOR_DARKGREY = '\033[90m'

    CONTROL_END = '\033[0m'

    MOVE_BACK = '\033[1D'
    MOVE_FORWARD = '\033[1C'

    ESCAPE_CODES = [COLOR_BLUE, COLOR_GREEN, COLOR_YELLOW, COLOR_RED,
                    COLOR_DARKGREY, CONTROL_END, MOVE_BACK, MOVE_FORWARD]

    """
    Class to help applications to colorize their outputs for terminals.

    This will probe the current terminal and colorize output only if the
    stdout is in a tty or the terminal type is recognized.
    """

    def __init__(self):
        self.HEADER = self.COLOR_BLUE
        self.PASS = self.COLOR_GREEN
        self.SKIP = self.COLOR_YELLOW
        self.FAIL = self.COLOR_RED
        self.INTERRUPT = self.COLOR_RED
        self.ERROR = self.COLOR_RED
        self.WARN = self.COLOR_YELLOW
        self.CANCEL = self.COLOR_YELLOW
        self.PARTIAL = self.COLOR_YELLOW
        self.ENDC = self.CONTROL_END
        self.LOWLIGHT = self.COLOR_DARKGREY
        self.enabled = True
        allowed_terms = ['linux', 'xterm', 'xterm-256color', 'vt100', 'screen',
                         'screen-256color', 'screen.xterm-256color']
        term = os.environ.get("TERM")
        config = settings.as_dict()
        colored = config.get('runner.output.colored')
        force_color = config.get('runner.output.color')
        if force_color == "never":
            self.disable()
        elif force_color == "auto":
            if not colored or not os.isatty(1) or term not in allowed_terms:
                self.disable()
        elif force_color != "always":
            raise ValueError("The value for runner.output.color must be one of "
                             "'always', 'never', 'auto' and not " + force_color)

    def disable(self):
        """
        Disable colors from the strings output by this class.
        """
        self.enabled = False
        self.HEADER = ''
        self.PASS = ''
        self.SKIP = ''
        self.FAIL = ''
        self.INTERRUPT = ''
        self.ERROR = ''
        self.WARN = ''
        self.CANCEL = ''
        self.PARTIAL = ''
        self.ENDC = ''
        self.LOWLIGHT = ''
        self.MOVE_BACK = ''
        self.MOVE_FORWARD = ''

    def header_str(self, msg):
        """
        Print a header string (blue colored).

        If the output does not support colors, just return the original string.
        """
        return self.HEADER + msg + self.ENDC

    def fail_header_str(self, msg):
        """
        Print a fail header string (red colored).

        If the output does not support colors, just return the original string.
        """
        return self.FAIL + msg + self.ENDC

    def warn_header_str(self, msg):
        """
        Print a warning header string (yellow colored).

        If the output does not support colors, just return the original string.
        """
        return self.WARN + msg + self.ENDC

    def healthy_str(self, msg):
        """
        Print a healthy string (green colored).

        If the output does not support colors, just return the original string.
        """
        return self.PASS + msg + self.ENDC

    def partial_str(self, msg):
        """
        Print a string that denotes partial progress (yellow colored).

        If the output does not support colors, just return the original string.
        """
        return self.PARTIAL + msg + self.ENDC

    def pass_str(self, msg='PASS', move=MOVE_BACK):
        """
        Print a pass string (green colored).

        If the output does not support colors, just return the original string.
        """
        return move + self.PASS + msg + self.ENDC

    def skip_str(self, msg='SKIP', move=MOVE_BACK):
        """
        Print a skip string (yellow colored).

        If the output does not support colors, just return the original string.
        """
        return move + self.SKIP + msg + self.ENDC

    def fail_str(self, msg='FAIL', move=MOVE_BACK):
        """
        Print a fail string (red colored).

        If the output does not support colors, just return the original string.
        """
        return move + self.FAIL + msg + self.ENDC

    def error_str(self, msg='ERROR', move=MOVE_BACK):
        """
        Print a error string (red colored).

        If the output does not support colors, just return the original string.
        """
        return move + self.ERROR + msg + self.ENDC

    def interrupt_str(self, msg='INTERRUPT', move=MOVE_BACK):
        """
        Print an interrupt string (red colored).

        If the output does not support colors, just return the original string.
        """
        return move + self.INTERRUPT + msg + self.ENDC

    def warn_str(self, msg='WARN', move=MOVE_BACK):
        """
        Print an warning string (yellow colored).

        If the output does not support colors, just return the original string.
        """
        return move + self.WARN + msg + self.ENDC


#: Transparently handles colored terminal, when one is used
TERM_SUPPORT = TermSupport()


#: A collection of mapping from test statuses to colors to be used
#: consistently across the various plugins
TEST_STATUS_MAPPING = {'PASS': TERM_SUPPORT.PASS,
                       'ERROR': TERM_SUPPORT.ERROR,
                       'FAIL': TERM_SUPPORT.FAIL,
                       'SKIP': TERM_SUPPORT.SKIP,
                       'WARN': TERM_SUPPORT.WARN,
                       'INTERRUPTED': TERM_SUPPORT.INTERRUPT,
                       'CANCEL': TERM_SUPPORT.CANCEL}


#: A collection of mapping from test status to formatting functions
#: to be used consistently across the various plugins
TEST_STATUS_DECORATOR_MAPPING = {'PASS': TERM_SUPPORT.pass_str,
                                 'ERROR': TERM_SUPPORT.error_str,
                                 'FAIL': TERM_SUPPORT.fail_str,
                                 'SKIP': TERM_SUPPORT.skip_str,
                                 'WARN': TERM_SUPPORT.warn_str,
                                 'INTERRUPTED': TERM_SUPPORT.interrupt_str,
                                 'CANCEL': TERM_SUPPORT.skip_str}


class _StdOutputFile:

    """
    File-like object which stores (_is_stdout, content) into the provided list
    """

    def __init__(self, is_stdout, records):
        """
        :param is_stdout: Is this output stdout or stderr
        :param records: list to store (is_stdout, written_message) records
        """
        self._is_stdout = is_stdout
        self._records = records

    def write(self, msg):
        """
        Record the message
        """
        self._records.append((self._is_stdout, msg))

    def writelines(self, iterable):
        """
        Record all messages
        """
        for line in iterable:
            self.write(line)

    def close(self):
        """ File-object methods """

    def flush(self):
        """ File-object methods """

    def isatty(self):
        """ File-object methods """
        return False

    def seek(self):
        """ File-object methods """

    def tell(self):
        """ File-object methods """

    def getvalue(self):
        """
        Get all messages written to this "file"
        """
        return "\n".join((_[1] for _ in self._records
                          if _[0] == self._is_stdout))


class StdOutput:

    """
    Class to modify sys.stdout/sys.stderr
    """
    #: List of records of stored output when stdout/stderr is disabled
    records = []

    def __init__(self):
        self.stdout = self._stdout = sys.stdout
        self.stderr = self._stderr = sys.stderr
        self.__configured = False

    def _paginator_in_use(self):
        """
        :return: True when we output into paginator
        """
        return bool(isinstance(sys.stdout, Paginator))

    @property
    def configured(self):
        """
        Determines if a configuration of any sort has been performed
        """
        return self.__configured

    def print_records(self):
        """
        Prints all stored messages as they occurred into streams they were
        produced for.
        """
        try:
            for stream, msg in self.records:
                if stream:
                    sys.stdout.write(msg)
                else:
                    sys.stderr.write(msg)
            del self.records[:]
        # IOError due to EPIPE is ignored.  That is to avoid having to
        # handle all IOErrors at the main application loop
        # indiscriminately.  By handling them here, we can be sure
        # that the failure was due to stdout or stderr not being
        # connected to an open PIPE.
        except IOError as detail:
            if not detail.errno == errno.EPIPE:
                raise

    def fake_outputs(self):
        """
        Replace sys.stdout/sys.stderr with in-memory-objects
        """
        sys.stdout = _StdOutputFile(True, self.records)
        sys.stderr = _StdOutputFile(False, self.records)
        self.__configured = True

    def enable_outputs(self):
        """
        Enable sys.stdout/sys.stderr (either with 2 streams or with paginator)
        """
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        self.__configured = True

    def enable_paginator(self):
        """
        Enable paginator
        """
        try:
            paginator = Paginator()
        except RuntimeError as details:
            # Paginator not available
            logging.getLogger('avocado.app.debug').error("Failed to enable "
                                                         "paginator: %s", details)
            return
        self.stdout = self.stderr = paginator
        self.__configured = True

    def enable_stderr(self):
        """
        Enable sys.stderr and disable sys.stdout
        """
        sys.stdout = open(os.devnull, 'w')
        sys.stderr = self.stderr
        self.__configured = True

    def close(self):
        """
        Enable original sys.stdout/sys.stderr and cleanup
        """
        paginator = None
        if self._paginator_in_use():
            paginator = sys.stdout
        self.enable_outputs()
        if paginator:
            paginator.close()


#: Allows modifying the sys.stdout/sys.stderr
STD_OUTPUT = StdOutput()


def early_start():
    """
    Replace all outputs with in-memory handlers
    """
    if os.environ.get('AVOCADO_LOG_DEBUG'):
        add_log_handler(LOG_UI.getChild("debug"), logging.StreamHandler,
                        sys.stdout, logging.DEBUG)
    if os.environ.get('AVOCADO_LOG_EARLY'):
        add_log_handler("", logging.StreamHandler, sys.stdout, logging.DEBUG)
        add_log_handler(LOG_JOB, logging.StreamHandler, sys.stdout,
                        logging.DEBUG)
    else:
        STD_OUTPUT.fake_outputs()
        add_log_handler("", MemStreamHandler, None, logging.DEBUG)
    logging.root.level = logging.DEBUG


CONFIG = []


def del_last_configuration():
    if len(CONFIG) == 1:
        return
    configuration = CONFIG.pop()
    for logger_name in configuration:
        disable_log_handler(logger_name)
    configuration = CONFIG[-1]
    for logger_name, handlers in configuration.items():
        logger = logging.getLogger(logger_name)
        for handler in handlers:
            logger.addHandler(handler)


def reconfigure(args):
    """
    Adjust logging handlers accordingly to app args and re-log messages.
    """
    def save_handler(logger_name, handler, configuration):
        if logger_name not in configuration:
            configuration[logger_name] = []
        configuration[logger_name].append(handler)

    # Delete last configuration
    if len(CONFIG) != 0:
        last_configuration = CONFIG[-1]
        for logger_name in last_configuration:
            disable_log_handler(logger_name)

    configuration = {}
    # Reconfigure stream loggers
    enabled = args.get("core.show")
    if not isinstance(enabled, list):
        enabled = ["app"]
        args["core.show"] = enabled
    if "none" in enabled:
        del enabled[:]
    elif "all" in enabled:
        enabled.extend([_ for _ in BUILTIN_STREAMS if _ not in enabled])
    if os.environ.get("AVOCADO_LOG_EARLY") and "early" not in enabled:
        enabled.append("early")
    if os.environ.get("AVOCADO_LOG_DEBUG") and "debug" not in enabled:
        enabled.append("debug")
    # TODO: Avocado relies on stdout/stderr on some places, re-log them here
    # for now. This should be removed once we replace them with logging.
    if enabled:
        if args.get('core.paginator') is True and TERM_SUPPORT.enabled:
            STD_OUTPUT.enable_paginator()
        STD_OUTPUT.enable_outputs()
    else:
        STD_OUTPUT.enable_stderr()
    STD_OUTPUT.print_records()
    if "app" in enabled:
        app_handler = ProgressStreamHandler()
        app_handler.setFormatter(logging.Formatter("%(message)s"))
        app_handler.addFilter(FilterInfoAndLess())
        app_handler.stream = STD_OUTPUT.stdout
        LOG_UI.addHandler(app_handler)
        LOG_UI.propagate = False
        LOG_UI.level = logging.DEBUG
        save_handler(LOG_UI.name, app_handler, configuration)
    else:
        disable_log_handler(LOG_UI)
    app_err_handler = ProgressStreamHandler()
    app_err_handler.setFormatter(logging.Formatter("%(message)s"))
    app_err_handler.addFilter(FilterWarnAndMore())
    app_err_handler.stream = STD_OUTPUT.stderr
    LOG_UI.addHandler(app_err_handler)
    LOG_UI.propagate = False
    save_handler(LOG_UI.name, app_err_handler, configuration)
    if not os.environ.get("AVOCADO_LOG_EARLY"):
        LOG_JOB.getChild("stdout").propagate = False
        LOG_JOB.getChild("stderr").propagate = False
        if "early" in enabled:
            handler = add_log_handler("", logging.StreamHandler,
                                      STD_OUTPUT.stdout, logging.DEBUG)
            save_handler("", handler, configuration)
            handler = add_log_handler(LOG_JOB, logging.StreamHandler,
                                      STD_OUTPUT.stdout, logging.DEBUG)
            save_handler(LOG_JOB.name, handler, configuration)
        else:
            disable_log_handler("")
    # Not enabled by env
    if not os.environ.get('AVOCADO_LOG_DEBUG'):
        if "debug" in enabled:
            handler = add_log_handler(LOG_UI.getChild("debug"),
                                      stream=STD_OUTPUT.stdout)
            save_handler(LOG_UI.getChild("debug").name, handler, configuration)
        else:
            disable_log_handler(LOG_UI.getChild("debug"))

    # Add custom loggers
    for name in [_ for _ in enabled if _ not in BUILTIN_STREAMS]:
        stream_level = re.split(r'(?<!\\):', name, maxsplit=1)
        name = stream_level[0]
        if len(stream_level) == 1:
            level = logging.DEBUG
        else:
            level = (int(stream_level[1]) if stream_level[1].isdigit()
                     else logging.getLevelName(stream_level[1].upper()))
        try:
            handler = add_log_handler(name, logging.StreamHandler,
                                      STD_OUTPUT.stdout, level)
            save_handler(name, handler, configuration)
        except ValueError as details:
            LOG_UI.error("Failed to set logger for --show %s:%s: %s.",
                         name, level, details)
            sys.exit(exit_codes.AVOCADO_FAIL)
    # Remove the in-memory handlers
    for handler in logging.root.handlers:
        if isinstance(handler, MemStreamHandler):
            logging.root.handlers.remove(handler)

    # Log early_messages
    for record in MemStreamHandler.log:
        logging.getLogger(record.name).handle(record)

    CONFIG.append(configuration)


class FilterWarnAndMore(logging.Filter):

    def filter(self, record):
        return record.levelno >= logging.WARN


class FilterInfoAndLess(logging.Filter):

    def filter(self, record):
        return record.levelno <= logging.INFO


class ProgressStreamHandler(logging.StreamHandler):

    """
    Handler class that allows users to skip new lines on each emission.
    """

    def emit(self, record):
        try:
            msg = self.format(record)
            if record.levelno < logging.INFO:   # Most messages are INFO
                pass
            elif record.levelno < logging.WARNING:
                msg = TERM_SUPPORT.header_str(msg)
            elif record.levelno < logging.ERROR:
                msg = TERM_SUPPORT.warn_header_str(msg)
            else:
                msg = TERM_SUPPORT.fail_header_str(msg)
            stream = self.stream
            skip_newline = False
            if hasattr(record, 'skip_newline'):
                skip_newline = record.skip_newline
            stream.write(msg)
            if not skip_newline:
                stream.write('\n')
            self.flush()
        except (KeyboardInterrupt, SystemExit):  # pylint: disable=W0706
            raise
        except Exception:  # pylint: disable=W0703
            self.handleError(record)


class MemStreamHandler(logging.StreamHandler):

    """
    Handler that stores all records in self.log (shared in all instances)
    """

    log = []

    def emit(self, record):
        self.log.append(record)

    def flush(self):
        """
        This is in-mem object, it does not require flushing
        """


class Paginator:

    """
    Paginator that uses less to display contents on the terminal.

    Contains cleanup handling for when user presses 'q' (to quit less).
    """

    def __init__(self):
        self.pipe = None
        paginator = os.environ.get('PAGER')
        if not paginator:
            try:
                paginator = "%s -FRX" % utils_path.find_command('less')
            except utils_path.CmdNotFoundError as details:
                raise RuntimeError("Unable to enable pagination: %s" % details)

        self.pipe = os.popen(paginator, 'w')

    def __del__(self):
        self.close()

    def close(self):
        if self.pipe:
            try:
                self.pipe.close()
            except OSError:
                pass

    def write(self, msg):
        if self.pipe:
            try:
                self.pipe.write(msg)
            except (OSError, ValueError):
                pass

    def flush(self):
        if self.pipe and not self.pipe.closed:
            self.pipe.flush()


def add_log_handler(logger, klass=logging.StreamHandler, stream=sys.stdout,
                    level=logging.INFO, fmt='%(name)s: %(message)s'):
    """
    Add handler to a logger.

    :param logger_name: the name of a :class:`logging.Logger` instance, that
                        is, the parameter to :func:`logging.getLogger`
    :param klass: Handler class (defaults to :class:`logging.StreamHandler`)
    :param stream: Logging stream, to be passed as an argument to ``klass``
                   (defaults to ``sys.stdout``)
    :param level: Log level (defaults to `INFO``)
    :param fmt: Logging format (defaults to ``%(name)s: %(message)s``)
    """
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    handler = klass(stream)
    handler.setLevel(level)
    if isinstance(fmt, str):
        fmt = logging.Formatter(fmt=fmt)
    handler.setFormatter(fmt)
    logger.addHandler(handler)
    logger.propagate = False
    return handler


def disable_log_handler(logger):
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    # Handlers might be reused elsewhere, can't delete them
    while logger.handlers:
        logger.handlers.pop()
    logger.handlers.append(logging.NullHandler())
    logger.propagate = False


class LoggingFile:

    """
    File-like object that will receive messages pass them to logging.
    """

    def __init__(self, prefixes=None, level=logging.DEBUG,
                 loggers=None):
        """
        Constructor. Sets prefixes and which loggers are going to be used.

        :param prefixes: Prefix per logger to be prefixed to each line.
        :param level: Log level to be used when writing messages.
        :param loggers: Loggers into which write should be issued. (list)
        """
        if not loggers:
            loggers = [logging.getLogger()]
        self._level = level
        self._loggers = loggers
        if prefixes is None:
            prefixes = [""] * len(loggers)
        self._prefixes = prefixes

    def write(self, data):
        """"
        Splits the line to individual lines and forwards them into loggers
        with expected prefixes. It includes the tailing newline <lf> as well
        as the last partial message. Do configure your logging to not to add
        newline <lf> automatically.
        :param data - Raw data (a string) that will be processed.
        """
        # splitlines() discards a trailing blank line, so use split() instead
        data_lines = data.split('\n')
        if len(data_lines) > 1:     # when not last line, contains \n
            self._log_line("%s\n" % data_lines[0])
        for line in data_lines[1:-1]:
            self._log_line("%s\n" % line)
        if data_lines[-1]:  # Last line does not contain \n
            self._log_line(data_lines[-1])

    def _log_line(self, line):
        """
        Forwards line to all the expected loggers along with expected prefix
        """
        for logger, prefix in zip(self._loggers, self._prefixes):
            logger.log(self._level, prefix + line)

    def flush(self):
        pass

    def isatty(self):
        return False

    def add_logger(self, logger, prefix=""):
        self._loggers.append(logger)
        self._prefixes.append(prefix)

    def rm_logger(self, logger):
        idx = self._loggers.index(logger)
        self._loggers.remove(logger)
        self._prefixes = self._prefixes[:idx] + self._prefixes[idx+1:]


class Throbber:

    """
    Produces a spinner used to notify progress in the application UI.
    """
    STEPS = ['-', '\\', '|', '/']
    # Only print a throbber when we're on a terminal
    if TERM_SUPPORT.enabled:
        MOVES = [TERM_SUPPORT.MOVE_BACK + STEPS[0],
                 TERM_SUPPORT.MOVE_BACK + STEPS[1],
                 TERM_SUPPORT.MOVE_BACK + STEPS[2],
                 TERM_SUPPORT.MOVE_BACK + STEPS[3]]
    else:
        MOVES = ['', '', '', '']

    def __init__(self):
        self.position = 0

    def _update_position(self):
        if self.position == (len(self.MOVES) - 1):
            self.position = 0
        else:
            self.position += 1

    def render(self):
        result = self.MOVES[self.position]
        self._update_position()
        return result


def log_plugin_failures(failures):
    """
    Log in the application UI failures to load a set of plugins

    :param failures: a list of load failures, usually coming from a
                     :class:`avocado.core.dispatcher.Dispatcher`
                     attribute `load_failures`
    """
    msg_fmt = 'Failed to load plugin from module "%s": %s :\n%s'
    config = settings.as_dict()
    silenced = config.get('plugins.skip_broken_plugin_notification')
    for failure in failures:
        if failure[0].module_name in silenced:
            continue
        if hasattr(failure[1], "__traceback__"):
            str_tb = ''.join(traceback.format_tb(failure[1].__traceback__))
        else:
            str_tb = "Traceback not available"
        LOG_UI.error(msg_fmt, failure[0].module_name,
                     failure[1].__repr__(), str_tb)



import json
import os
import signal
import stat
import subprocess
import time
import unittest

from avocado.core import exit_codes
from avocado.utils import process, script

from .. import (AVOCADO, BASEDIR, TestCaseTmpDir, skipOnLevelsInferiorThan,
                skipUnlessPathExists)

AVOCADO_TEST_OK = """#!/usr/bin/env python
from avocado import Test
from avocado import main

class PassTest(Test):
    def test(self):
        pass

if __name__ == "__main__":
    main()
"""


AVOCADO_TEST_SLEEP_ELEVEN = """#!/usr/bin/env python
import time

from avocado import Test
from avocado import main

class SleepEleven(Test):
    def test(self):
        time.sleep(10)
    def test_2(self):
        time.sleep(1)

time.sleep(11)

if __name__ == "__main__":
    main()
"""


AVOCADO_TEST_MULTIPLE_CLASSES = """#!/usr/bin/env python
import time

from avocado import Test
from avocado import main

class First(Test):
    def test(self):
        pass

class Second(Test):
    def test(self):
        pass

if __name__ == "__main__":
    main()
"""

AVOCADO_TEST_MULTIPLE_METHODS_SAME_NAME = """#!/usr/bin/env python
from avocado import Test
from avocado import main

class Multiple(Test):
    def test(self):
        raise

    def test(self):
        pass

if __name__ == "__main__":
    main()
"""

NOT_A_TEST = """
def hello():
    print('Hello World!')
"""

PY_SIMPLE_TEST = """#!/usr/bin/env python
def hello():
    print('Hello World!')

if __name__ == "__main__":
    hello()
"""

SIMPLE_TEST = """#!/bin/sh
true
"""

AVOCADO_SIMPLE_PYTHON_LIKE_MULTIPLE_FILES = """#!/usr/bin/env python
# A simple test (executable bit set when saved to file) that looks like
# an Avocado instrumented test, with base class on separate file
from avocado import Test
from avocado import main
from test2 import *

class BasicTestSuite(SuperTest):
    '''
    :avocado: disable
    '''

    def test1(self):
        self.xxx()
        self.assertTrue(True)

if __name__ == '__main__':
    main()
"""

AVOCADO_SIMPLE_PYTHON_LIKE_MULTIPLE_FILES_LIB = """
#!/usr/bin/python

from avocado import Test

class SuperTest(Test):
    def xxx(self):
        print "ahoj"
"""

AVOCADO_TEST_SIMPLE_USING_MAIN = """#!/usr/bin/env python
from avocado import main

if __name__ == "__main__":
    main()
"""


class LoaderTestFunctional(TestCaseTmpDir):

    MODE_0664 = (stat.S_IRUSR | stat.S_IWUSR |
                 stat.S_IRGRP | stat.S_IWGRP |
                 stat.S_IROTH)

    MODE_0775 = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
                 stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP |
                 stat.S_IROTH | stat.S_IXOTH)

    def _test(self, name, content, exp_str, mode=MODE_0664, count=1):
        test_script = script.TemporaryScript(name, content,
                                             'avocado_loader_test',
                                             mode=mode)
        test_script.save()
        cmd_line = ('%s -V list %s' % (AVOCADO, test_script.path))
        result = process.run(cmd_line)
        self.assertIn('%s: %s' % (exp_str, count), result.stdout_text)
        test_script.remove()

    def _run_with_timeout(self, cmd_line, timeout):
        current_time = time.time()
        deadline = current_time + timeout
        test_process = subprocess.Popen(cmd_line, stdout=subprocess.PIPE,  # pylint: disable=W1509
                                        stderr=subprocess.PIPE,
                                        preexec_fn=os.setsid, shell=True)
        while not test_process.poll():
            if time.time() > deadline:
                os.killpg(os.getpgid(test_process.pid), signal.SIGKILL)
                self.fail("Failed to run test under %s seconds" % timeout)
            time.sleep(0.05)
        self.assertEqual(test_process.returncode, exit_codes.AVOCADO_TESTS_FAIL)

    def test_simple(self):
        self._test('simpletest.sh', SIMPLE_TEST, 'simple', self.MODE_0775)

    def test_simple_not_exec(self):
        # 2 because both FileLoader and the TAP loader cannot recognize the test
        self._test('simpletest.sh', SIMPLE_TEST, 'not_a_test', count=2)

    def test_pass(self):
        self._test('passtest.py', AVOCADO_TEST_OK, 'instrumented')

    def test_not_python_module(self):
        # 2 because both FileLoader and the TAP loader cannot recognize the test
        self._test('passtest', AVOCADO_TEST_OK, 'not_a_test', count=2)

    @skipOnLevelsInferiorThan(2)
    def test_sleep_a_lot(self):
        """
        Verifies that the test loader, at list time, does not load the Python
        module and thus executes its contents.

        :avocado: tags=parallel:1
        """
        test_script = script.TemporaryScript('sleepeleven.py',
                                             AVOCADO_TEST_SLEEP_ELEVEN,
                                             'avocado_loader_test',
                                             mode=self.MODE_0664)
        test_script.save()
        cmd_line = ('%s -V list %s' % (AVOCADO, test_script.path))
        initial_time = time.time()
        result = process.run(cmd_line, ignore_status=True)
        test_script.remove()
        actual_time = time.time() - initial_time
        self.assertLess(actual_time, 3.0,
                        ("Took more than 3 seconds to list tests. Loader "
                         "probably loaded/executed Python code and slept for "
                         "eleven seconds."))
        self.assertIn(b'instrumented: 2', result.stdout)

    def test_multiple_class(self):
        self._test('multipleclasses.py', AVOCADO_TEST_MULTIPLE_CLASSES,
                   'instrumented', self.MODE_0664, 2)

    def test_multiple_methods_same_name(self):
        self._test('multiplemethods.py', AVOCADO_TEST_MULTIPLE_METHODS_SAME_NAME,
                   'instrumented', self.MODE_0664, 1)

    def test_load_not_a_test(self):
        self._test('notatest.py', NOT_A_TEST, 'simple', self.MODE_0775)

    def test_load_not_a_test_not_exec(self):
        # 2 because both FileLoader and the TAP loader cannot recognize the test
        self._test('notatest.py', NOT_A_TEST, 'not_a_test', count=2)

    @skipOnLevelsInferiorThan(2)
    def test_runner_simple_python_like_multiple_files(self):
        """
        :avocado: tags=parallel:1
        """
        mylib = script.TemporaryScript(
            'test2.py',
            AVOCADO_SIMPLE_PYTHON_LIKE_MULTIPLE_FILES_LIB,
            'avocado_simpletest_functional',
            self.MODE_0664)
        mylib.save()
        mytest = script.Script(
            os.path.join(os.path.dirname(mylib.path), 'test.py'),
            AVOCADO_SIMPLE_PYTHON_LIKE_MULTIPLE_FILES)
        os.chdir(BASEDIR)
        mytest.save()
        cmd_line = "%s -V list %s" % (AVOCADO, mytest)
        result = process.run(cmd_line)
        self.assertIn(b'simple: 1', result.stdout)
        # job should be able to finish under 5 seconds. If this fails, it's
        # possible that we hit the "simple test fork bomb" bug
        cmd_line = ("%s run --disable-sysinfo --job-results-dir '%s' -- '%s'"
                    % (AVOCADO, self.tmpdir.name, mytest))
        self._run_with_timeout(cmd_line, 5)

    @skipOnLevelsInferiorThan(2)
    def test_simple_using_main(self):
        """
        :avocado: tags=parallel:1
        """
        mytest = script.TemporaryScript("simple_using_main.py",
                                        AVOCADO_TEST_SIMPLE_USING_MAIN,
                                        'avocado_simpletest_functional')
        mytest.save()
        os.chdir(BASEDIR)
        # job should be able to finish under 5 seconds. If this fails, it's
        # possible that we hit the "simple test fork bomb" bug
        cmd_line = ("%s run --disable-sysinfo --job-results-dir '%s' -- '%s'"
                    % (AVOCADO, self.tmpdir.name, mytest))
        self._run_with_timeout(cmd_line, 5)

    def test_python_unittest(self):
        test_path = os.path.join(BASEDIR, "selftests", ".data", "unittests.py")
        cmd = ("%s run --disable-sysinfo --job-results-dir %s --json - -- %s"
               % (AVOCADO, self.tmpdir.name, test_path))
        result = process.run(cmd, ignore_status=True)
        jres = json.loads(result.stdout_text)
        self.assertEqual(result.exit_status, 1, result)
        exps = [("unittests.Second.test_fail", "FAIL"),
                ("unittests.Second.test_error", "ERROR"),
                ("unittests.Second.test_skip", "CANCEL"),
                ("unittests.First.test_pass", "PASS")]
        for test in jres["tests"]:
            for exp in exps:
                if exp[0] in test["id"]:
                    self.assertEqual(test["status"], exp[1], "Status of %s not"
                                     " as expected\n%s" % (exp, result))
                    exps.remove(exp)
                    break
            else:
                self.fail("No expected result for %s\n%s\n\nexps = %s"
                          % (test["id"], result, exps))
        self.assertFalse(exps, "Some expected result not matched to actual"
                         "results:\n%s\n\nexps = %s" % (result, exps))

    def test_list_subtests_filter(self):
        """
        Check whether the subtests filter works for both INSTRUMENTED
        and SIMPLE in a directory list.
        """
        cmd = "%s list examples/tests/:fail" % AVOCADO
        result = process.run(cmd)
        expected = (b"INSTRUMENTED examples/tests/assert.py:Assert.test_fails_to_raise\n"
                    b"INSTRUMENTED examples/tests/doublefail.py:DoubleFail.test\n"
                    b"INSTRUMENTED examples/tests/fail_on_exception.py:FailOnException.test\n"
                    b"INSTRUMENTED examples/tests/failtest.py:FailTest.test\n"
                    b"SIMPLE       examples/tests/failtest.sh\n")
        self.assertEqual(expected, result.stdout)

    @skipUnlessPathExists('/bin/sh')
    def test_loader_and_external_runner_incompatibility(self):
        """
        Check if the user is inform about incompatibility between loader and
        external_runner.
        """
        test_script = script.TemporaryScript('simpletest.sh', SIMPLE_TEST,
                                             'avocado_loader_test',
                                             mode=self.MODE_0775)
        test_script.save()

        cmd = ("%s run --loaders=FOO "
               "--external-runner=/bin/sh %s") % (AVOCADO, test_script.path)
        result = process.run(cmd)
        expected_warning = ("The loaders and external-runner are incompatible."
                            "The values in loaders will be ignored.")
        self.assertIn(expected_warning, result.stderr_text)

        cmd = "%s run --external-runner=/bin/sh %s" % (AVOCADO,
                                                       test_script.path)
        result = process.run(cmd)
        self.assertNotIn(expected_warning, result.stderr_text)

        test_script.remove()


if __name__ == '__main__':
    unittest.main()



import glob
import os
import unittest
import xml.dom.minidom

from avocado.core import exit_codes
from avocado.utils import genio, process, script

from .. import AVOCADO, BASEDIR, TestCaseTmpDir, skipOnLevelsInferiorThan

SCRIPT_CONTENT = """#!/bin/bash
sleep 2
"""

PYTHON_CONTENT = """#!/usr/bin/env python
import time
from avocado import Test

class Dummy(Test):
    def test00sleep(self):
        time.sleep(10)
    def test01pass(self):
        pass
    def test02pass(self):
        pass
"""


class ParseXMLError(Exception):
    pass


class JobTimeOutTest(TestCaseTmpDir):

    def setUp(self):
        super(JobTimeOutTest, self).setUp()
        self.script = script.TemporaryScript(
            'sleep.sh',
            SCRIPT_CONTENT,
            'avocado_timeout_functional')
        self.script.save()
        self.py = script.TemporaryScript(
            'sleep_test.py',
            PYTHON_CONTENT,
            'avocado_timeout_functional')
        self.py.save()

    def run_and_check(self, cmd_line, e_rc, e_ntests, e_nerrors, e_nfailures,
                      e_nskip):
        os.chdir(BASEDIR)
        result = process.run(cmd_line, ignore_status=True)
        xml_output = result.stdout
        self.assertEqual(result.exit_status, e_rc,
                         "Avocado did not return rc %d:\n%s" %
                         (e_rc, result))
        try:
            xunit_doc = xml.dom.minidom.parseString(xml_output)
        except Exception as detail:
            raise ParseXMLError("Failed to parse content: %s\n%s" %
                                (detail, xml_output))

        testsuite_list = xunit_doc.getElementsByTagName('testsuite')
        self.assertEqual(len(testsuite_list), 1, 'More than one testsuite tag')

        testsuite_tag = testsuite_list[0]
        self.assertEqual(len(testsuite_tag.attributes), 7,
                         'The testsuite tag does not have 7 attributes. '
                         'XML:\n%s' % xml_output)

        n_tests = int(testsuite_tag.attributes['tests'].value)
        self.assertEqual(n_tests, e_ntests,
                         "Unexpected number of executed tests, "
                         "XML:\n%s" % xml_output)

        n_errors = int(testsuite_tag.attributes['errors'].value)
        self.assertEqual(n_errors, e_nerrors,
                         "Unexpected number of test errors, "
                         "XML:\n%s" % xml_output)

        n_failures = int(testsuite_tag.attributes['failures'].value)
        self.assertEqual(n_failures, e_nfailures,
                         "Unexpected number of test failures, "
                         "XML:\n%s" % xml_output)

        n_skip = int(testsuite_tag.attributes['skipped'].value)
        self.assertEqual(n_skip, e_nskip,
                         "Unexpected number of test skips, "
                         "XML:\n%s" % xml_output)

    def _check_timeout_msg(self, idx):
        res_dir = os.path.join(self.tmpdir.name, "latest", "test-results")
        debug_log_paths = glob.glob(os.path.join(res_dir, "%s-*" % idx, "debug.log"))
        debug_log = genio.read_file(debug_log_paths[0])
        self.assertIn("Runner error occurred: Timeout reached", debug_log,
                      "Runner error occurred: Timeout reached message not "
                      "in the %sst test's debug.log:\n%s"
                      % (idx, debug_log))
        self.assertIn("Traceback (most recent call last)", debug_log,
                      "Traceback not present in the %sst test's debug.log:\n%s"
                      % (idx, debug_log))

    @skipOnLevelsInferiorThan(1)
    def test_sleep_longer_timeout(self):
        """:avocado: tags=parallel:1"""
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--xunit - --job-timeout=5 %s examples/tests/passtest.py' %
                    (AVOCADO, self.tmpdir.name, self.script.path))
        self.run_and_check(cmd_line, 0, 2, 0, 0, 0)

    def test_sleep_short_timeout(self):
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--xunit - --job-timeout=1 %s examples/tests/passtest.py' %
                    (AVOCADO, self.tmpdir.name, self.script.path))
        self.run_and_check(cmd_line, exit_codes.AVOCADO_JOB_INTERRUPTED,
                           2, 1, 0, 1)
        self._check_timeout_msg(1)

    def test_sleep_short_timeout_with_test_methods(self):
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--xunit - --job-timeout=1 %s' %
                    (AVOCADO, self.tmpdir.name, self.py.path))
        self.run_and_check(cmd_line, exit_codes.AVOCADO_JOB_INTERRUPTED,
                           3, 1, 0, 2)
        self._check_timeout_msg(1)

    def test_invalid_values(self):
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--job-timeout=1,5 examples/tests/passtest.py'
                    % (AVOCADO, self.tmpdir.name))
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_FAIL)
        self.assertIn(b'invalid time_to_seconds value', result.stderr)
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--job-timeout=123x examples/tests/passtest.py'
                    % (AVOCADO, self.tmpdir.name))
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_FAIL)
        self.assertIn(b'invalid time_to_seconds', result.stderr)

    def test_valid_values(self):
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--job-timeout=123 examples/tests/passtest.py'
                    % (AVOCADO, self.tmpdir.name))
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_ALL_OK)
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--job-timeout=123s examples/tests/passtest.py'
                    % (AVOCADO, self.tmpdir.name))
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_ALL_OK)
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--job-timeout=123m examples/tests/passtest.py'
                    % (AVOCADO, self.tmpdir.name))
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_ALL_OK)
        cmd_line = ('%s run --job-results-dir %s --disable-sysinfo '
                    '--job-timeout=123h examples/tests/passtest.py'
                    % (AVOCADO, self.tmpdir.name))
        result = process.run(cmd_line, ignore_status=True)
        self.assertEqual(result.exit_status, exit_codes.AVOCADO_ALL_OK)

    def tearDown(self):
        super(JobTimeOutTest, self).tearDown()
        self.script.remove()
        self.py.remove()


if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2014-2016,2018
# Authors: Cleber Rosa <crosa@redhat.com>
#          Lukas Doktor <ldoktor@redhat.com>

"""
Safe (AST based) test loader module utilities
"""

import ast
import collections
import json
import os
import re
import sys
from importlib.machinery import PathFinder

from ..utils import data_structures


class PythonModule:
    """
    Representation of a Python module that might contain interesting classes

    By default, it uses module and class names that matches Avocado
    instrumented tests, but it's supposed to be agnostic enough to
    be used for, say, Python unittests.
    """
    __slots__ = ('path', 'klass_imports', 'mod_imports', 'mod', 'imported_objects',
                 'module', 'klass')

    def __init__(self, path, module='avocado', klass='Test'):
        """
        Instantiates a new PythonModule representation

        :param path: path to a Python source code file
        :type path: str
        :param module: the original module name from where the
                       possibly interesting class must have been
                       imported from
        :type module: str
        :param klass: the possibly interesting class original name
        :type klass: str
        """
        self.klass_imports = set()
        self.mod_imports = set()
        if os.path.isdir(path):
            path = os.path.join(path, "__init__.py")
        self.path = path
        self.module = module
        self.klass = klass
        # A dict that keeps track of objects names and importable entities
        #   key => object name from this module point of view
        #   value => Something-like a directory path to the import.
        #            Basically a $path/$module/$variable string, but depending
        #            on the type of import, it can be also be $path/$module.
        self.imported_objects = {}
        with open(self.path) as source_file:
            self.mod = ast.parse(source_file.read(), self.path)

    def is_matching_klass(self, klass):
        """
        Detect whether given class directly defines itself as <module>.<klass>

        It can either be a <klass> that inherits from a test "symbol", like:

        ```class FooTest(Test)```

        Or from an <module>.<klass> symbol, like in:

        ```class FooTest(avocado.Test)```

        :type klass: :class:`_ast.ClassDef`
        :rtype: bool
        """
        # Is it inherited from Test? 'class FooTest(Test):'
        if self.klass_imports:
            base_ids = [base.id for base in klass.bases
                        if isinstance(base, ast.Name)]
            # Looking for a 'class FooTest(Test):'
            if not self.klass_imports.isdisjoint(base_ids):
                return True

        # Is it inherited from avocado.Test? 'class FooTest(avocado.Test):'
        if self.mod_imports:
            for base in klass.bases:
                if not isinstance(base, ast.Attribute):
                    # Check only 'module.Class' bases
                    continue
                cls_module = base.value.id
                cls_name = base.attr
                if cls_module in self.mod_imports and cls_name == self.klass:
                    return True
        return False

    def add_imported_object(self, statement):
        """
        Keeps track of objects names and importable entities
        """
        path = os.path.abspath(os.path.dirname(self.path))
        if getattr(statement, 'module', None) is not None:
            module_path = statement.module.replace('.', os.path.sep)
            path = os.path.join(path, module_path)
        else:
            # Module has no name, its path is relative to the directory
            # structure
            level = getattr(statement, 'level', 0)
            for _ in range(level - 1):
                path = os.path.dirname(path)
        for name in statement.names:
            full_path = os.path.join(path, name.name.replace('.', os.path.sep))
            final_name = self._get_name_from_alias_statement(name)
            self.imported_objects[final_name] = full_path

    @staticmethod
    def _get_name_from_alias_statement(alias):
        """Returns the aliased name or original one."""
        return alias.asname if alias.asname else alias.name

    def _handle_import_from(self, statement):
        self.add_imported_object(statement)
        if statement.module != self.module:
            return
        name = statement_import_as(statement).get(self.klass, None)
        if name is not None:
            self.klass_imports.add(name)

    @staticmethod
    def _all_module_level_names(full_module_name):
        result = []
        components = full_module_name.split(".")[:-1]
        for topmost in range(len(components)):
            result.append(".".join(components[-topmost:]))
            components.pop()
        return result

    def _handle_import(self, statement):
        self.add_imported_object(statement)
        imported_as = statement_import_as(statement)
        name = imported_as.get(self.module, None)
        if name is not None:
            self.mod_imports.add(name)
        for as_name in imported_as.values():
            for mod_name in self._all_module_level_names(as_name):
                if mod_name == self.module:
                    self.mod_imports.add(mod_name)

    def iter_classes(self):
        """
        Iterate through classes and keep track of imported avocado statements
        """
        for statement in self.mod.body:
            # Looking for a 'from <module> import <klass>'
            if isinstance(statement, ast.ImportFrom):
                self._handle_import_from(statement)

            # Looking for a 'import <module>'
            elif isinstance(statement, ast.Import):
                self._handle_import(statement)

            # Looking for a 'class Anything(anything):'
            elif isinstance(statement, ast.ClassDef):
                yield statement


def statement_import_as(statement):
    """
    Returns a mapping of imported module names whether using aliases or not

    :param statement: an AST import statement
    :type statement: ast.Import
    :returns: a mapping of names {<realname>: <alias>} of modules imported
    :rtype: dict
    """
    result = {}
    for name in statement.names:
        if name.asname is not None:
            result[name.name] = name.asname
        else:
            result[name.name] = name.name
    return result


def modules_imported_as(module):
    """
    Returns a mapping of imported module names whether using aliases or not

    The goal of this utility function is to return the name of the import
    as used in the rest of the module, whether an aliased import was used
    or not.

    For code such as:

    >>> import foo as bar

    This function should return {"foo": "bar"}

    And for code such as:

    >>> import foo

    It should return {"foo": "foo"}

    Please note that only global level imports are looked at. If there are
    imports defined, say, inside functions or class definitions, they will
    not be seen by this function.

    :param module: module, as parsed by :func:`ast.parse`
    :type module: :class:`_ast.Module`
    :returns: a mapping of names {<realname>: <alias>} of modules imported
    :rtype: dict
    """
    result = {}
    for statement in module.body:
        if isinstance(statement, ast.Import):
            result.update(statement_import_as(statement))
    return result


#: Gets the docstring directive value from a string. Used to tweak
#: test behavior in various ways
DOCSTRING_DIRECTIVE_RE_RAW = r'\s*:avocado:[ \t]+(([a-zA-Z0-9]+?[a-zA-Z0-9_:,\=\-\.]*)|(r[a-zA-Z0-9]+?[a-zA-Z0-9_:,\=\{\}\"\-\.\/ ]*))\s*$'
# the RE will match `:avocado: tags=category` or `:avocado: requirements={}`
DOCSTRING_DIRECTIVE_RE = re.compile(DOCSTRING_DIRECTIVE_RE_RAW)


def get_docstring_directives(docstring):
    """
    Returns the values of the avocado docstring directives

    :param docstring: the complete text used as documentation
    :type docstring: str

    :rtype: builtin.list
    """
    result = []
    if docstring is None:
        return result
    for line in docstring.splitlines():
        try:
            match = DOCSTRING_DIRECTIVE_RE.match(line)
            if match:
                result.append(match.groups()[0])
        except TypeError:
            pass
    return result


def check_docstring_directive(docstring, directive):
    """
    Checks if there's a given directive in a given docstring

    :rtype: bool
    """
    return directive in get_docstring_directives(docstring)


def get_docstring_directives_tags(docstring):
    """
    Returns the test categories based on a `:avocado: tags=category`
    docstring

    :rtype: dict
    """
    result = {}
    for item in get_docstring_directives(docstring):
        if item.startswith('tags='):
            _, comma_tags = item.split('tags=', 1)
            for tag in comma_tags.split(','):
                if not tag:
                    continue
                if ':' in tag:
                    key, val = tag.split(':', 1)
                    if key in result:
                        result[key].add(val)
                    else:
                        result[key] = set([val])
                else:
                    result[tag] = None
    return result


def get_docstring_directives_requirements(docstring):
    """
    Returns the test requirements from docstring patterns like
    `:avocado: requirement={}`.

    :rtype: list
    """
    requirements = []
    for item in get_docstring_directives(docstring):
        if item.startswith('requirement='):
            _, requirement_str = item.split('requirement=', 1)
            try:
                requirements.append(json.loads(requirement_str))
            except json.decoder.JSONDecodeError:
                # ignore requirement in case of malformed dictionary
                continue
    return requirements


def find_class_and_methods(path, method_pattern=None, base_class=None):
    """
    Attempts to find methods names from a given Python source file

    :param path: path to a Python source code file
    :type path: str
    :param method_pattern: compiled regex to match against method name
    :param base_class: only consider classes that inherit from a given
                       base class (or classes that inherit from any class
                       if None is given)
    :type base_class: str or None
    :returns: an ordered dictionary with classes as keys and methods as values
    :rtype: collections.OrderedDict
    """
    def inherits_from_base_class(class_statement, base_class_name):
        base_ids = [base.id for base in class_statement.bases
                    if hasattr(base, 'id')]
        return base_class_name in base_ids

    result = collections.OrderedDict()
    with open(path) as source_file:
        mod = ast.parse(source_file.read(), path)

    for statement in mod.body:
        if isinstance(statement, ast.ClassDef):
            if base_class is not None:
                if not inherits_from_base_class(statement,
                                                base_class):
                    continue
            if method_pattern is None:
                methods = [st.name for st in statement.body if
                           isinstance(st, ast.FunctionDef)]
                methods = data_structures.ordered_list_unique(methods)
            else:
                methods = [st.name for st in statement.body if
                           isinstance(st, ast.FunctionDef) and
                           method_pattern.match(st.name)]
                methods = data_structures.ordered_list_unique(methods)
            result[statement.name] = methods
    return result


def get_methods_info(statement_body, class_tags, class_requirements):
    """
    Returns information on an Avocado instrumented test method
    """
    methods_info = []
    for st in statement_body:
        if (isinstance(st, ast.FunctionDef) and
                st.name.startswith('test')):
            docstring = ast.get_docstring(st)

            mt_tags = get_docstring_directives_tags(docstring)
            mt_tags.update(class_tags)

            mt_requirements = get_docstring_directives_requirements(docstring)
            mt_requirements.extend(class_requirements)

            methods = [method for method, _, _ in methods_info]
            if st.name not in methods:
                methods_info.append((st.name, mt_tags, mt_requirements))

    return methods_info


def _determine_match_avocado(module, klass, docstring):
    """
    Implements the match check for Avocado Instrumented Tests
    """
    directives = get_docstring_directives(docstring)
    if 'disable' in directives:
        return True
    if 'enable' in directives:
        return True
    if 'recursive' in directives:
        return True
    # Still not decided, try inheritance
    return module.is_matching_klass(klass)


def _extend_test_list(current, new):
    for test in new:
        test_method_name = test[0]
        if test_method_name not in [_[0] for _ in current]:
            current.append(test)


def _examine_class(path, class_name, match, target_module, target_class,
                   determine_match):
    """
    Examine a class from a given path

    :param path: path to a Python source code file
    :type path: str
    :param class_name: the specific class to be found
    :type path: str
    :param match: whether the inheritance from <target_module.target_class> has
                  been determined or not
    :type match: bool
    :param target_module: the module name under which the target_class lives
    :type target_module: str
    :param target_class: the name of the class that class_name should
                         ultimately inherit from
    :type target_class: str
    :param determine_match: a callable that will determine if a match has
                            occurred or not
    :type determine_match: function
    :returns: tuple where first item is a list of test methods detected
              for given class; second item is set of class names which
              look like avocado tests but are force-disabled.
    :rtype: tuple
    """
    module = PythonModule(path, target_module, target_class)
    info = []
    disabled = set()

    for klass in module.iter_classes():
        if class_name != klass.name:
            continue

        docstring = ast.get_docstring(klass)

        if match is False:
            match = determine_match(module, klass, docstring)

        info = get_methods_info(klass.body,
                                get_docstring_directives_tags(docstring),
                                get_docstring_directives_requirements(
                                    docstring))

        # Getting the list of parents of the current class
        parents = klass.bases

        # From this point we use `_$variable` to name temporary returns
        # from method calls that are to-be-assigned/combined with the
        # existing `$variable`.

        # Searching the parents in the same module
        for parent in parents[:]:
            # Looking for a 'class FooTest(Parent)'
            if not isinstance(parent, ast.Name):
                # 'class FooTest(bar.Bar)' not supported withing
                # a module
                continue
            parent_class = parent.id
            _info, _disabled, _match = _examine_class(module.path, parent_class,
                                                      match, target_module,
                                                      target_class,
                                                      determine_match)
            if _info:
                parents.remove(parent)
                _extend_test_list(info, _info)
                disabled.update(_disabled)
            if _match is not match:
                match = _match

        # If there are parents left to be discovered, they
        # might be in a different module.
        for parent in parents:
            if hasattr(parent, 'value'):
                if hasattr(parent.value, 'id'):
                    # We know 'parent.Class' or 'asparent.Class' and need
                    # to get path and original_module_name. Class is given
                    # by parent definition.
                    _parent = module.imported_objects.get(parent.value.id)
                    if _parent is None:
                        # We can't examine this parent (probably broken
                        # module)
                        continue
                    parent_path = os.path.dirname(_parent)
                    parent_module = os.path.basename(_parent)
                    parent_class = parent.attr
                else:
                    # We don't support multi-level 'parent.parent.Class'
                    continue
            else:
                # We only know 'Class' or 'AsClass' and need to get
                # path, module and original class_name
                _parent = module.imported_objects.get(parent.id)
                if _parent is None:
                    # We can't examine this parent (probably broken
                    # module)
                    continue
                parent_path, parent_module, parent_class = (
                    _parent.rsplit(os.path.sep, 2))

            modules_paths = [parent_path,
                             os.path.dirname(module.path)] + sys.path
            found_spec = PathFinder.find_spec(parent_module, modules_paths)
            if found_spec is not None:
                _info, _disabled, _match = _examine_class(found_spec.origin,
                                                          parent_class,
                                                          match,
                                                          target_module,
                                                          target_class,
                                                          _determine_match_avocado)
            if _info:
                _extend_test_list(info, _info)
                disabled.update(_disabled)
            if _match is not match:
                match = _match

    return info, disabled, match


def find_python_tests(module_name, class_name, determine_match, path):
    """
    Attempts to find Python tests from source files

    A Python test in this context is a method within a specific type
    of class (or that inherits from a specific class).

    :param module_name: the name of the module from which a class should
                        have come from
    :type module_name: str
    :param class_name: the name of the class that is considered to contain
                       test methods
    :type class_name: str
    :type determine_match: a callable that will determine if a given module
                           and class is contains valid Python tests
    :type determine_match: function
    :param path: path to a Python source code file
    :type path: str
    :returns: tuple where first item is dict with class name and additional
              info such as method names and tags; the second item is
              set of class names which look like Python tests but have been
              forcefully disabled.
    :rtype: tuple
    """
    module = PythonModule(path, module_name, class_name)
    # The resulting test classes
    result = collections.OrderedDict()
    disabled = set()

    for klass in module.iter_classes():
        docstring = ast.get_docstring(klass)
        # Looking for a class that has in the docstring either
        # ":avocado: enable" or ":avocado: disable
        if check_docstring_directive(docstring, 'disable'):
            disabled.add(klass.name)
            continue

        if check_docstring_directive(docstring, 'enable'):
            info = get_methods_info(klass.body,
                                    get_docstring_directives_tags(docstring),
                                    get_docstring_directives_requirements(
                                        docstring))
            result[klass.name] = info
            continue

        # From this point onwards we want to do recursive discovery, but
        # for now we don't know whether it is avocado.Test inherited
        # (Ifs are optimized for readability, not speed)

        # If "recursive" tag is specified, it is forced as test
        if check_docstring_directive(docstring, 'recursive'):
            is_valid_test = True
        else:
            is_valid_test = module.is_matching_klass(klass)
        info = get_methods_info(klass.body,
                                get_docstring_directives_tags(docstring),
                                get_docstring_directives_requirements(
                                    docstring))
        _disabled = set()

        # Getting the list of parents of the current class
        parents = klass.bases

        # Searching the parents in the same module
        for parent in parents[:]:
            # Looking for a 'class FooTest(Parent)'
            if not isinstance(parent, ast.Name):
                # 'class FooTest(bar.Bar)' not supported withing
                # a module
                continue
            parent_class = parent.id
            _info, _dis, _python_test = _examine_class(module.path,
                                                       parent_class,
                                                       is_valid_test,
                                                       module_name,
                                                       class_name,
                                                       determine_match)
            if _info:
                parents.remove(parent)
                _extend_test_list(info, _info)
                _disabled.update(_dis)
            if _python_test is not is_valid_test:
                is_valid_test = _python_test

        # If there are parents left to be discovered, they
        # might be in a different module.
        for parent in parents:
            if hasattr(parent, 'value'):
                if hasattr(parent.value, 'id'):
                    # We know 'parent.Class' or 'asparent.Class' and need
                    # to get path and original_module_name. Class is given
                    # by parent definition.
                    _parent = module.imported_objects.get(parent.value.id)
                    if _parent is None:
                        # We can't examine this parent (probably broken
                        # module)
                        continue
                    parent_path = os.path.dirname(_parent)
                    parent_module = os.path.basename(_parent)
                    parent_class = parent.attr
                else:
                    # We don't support multi-level 'parent.parent.Class'
                    continue
            else:
                # We only know 'Class' or 'AsClass' and need to get
                # path, module and original class_name
                _parent = module.imported_objects.get(parent.id)
                if _parent is None:
                    # We can't examine this parent (probably broken
                    # module)
                    continue
                parent_path, parent_module, parent_class = (
                    _parent.rsplit(os.path.sep, 2))

            modules_paths = [parent_path,
                             os.path.dirname(module.path)] + sys.path
            found_spec = PathFinder.find_spec(parent_module, modules_paths)
            if found_spec is None:
                continue
            _info, _dis, _python_test = _examine_class(found_spec.origin,
                                                       parent_class,
                                                       is_valid_test,
                                                       module_name,
                                                       class_name,
                                                       determine_match)
            if _info:
                info.extend(_info)
                _disabled.update(_dis)
            if _python_test is not is_valid_test:
                is_valid_test = _python_test

        # Only update the results if this was detected as 'avocado.Test'
        if is_valid_test:
            result[klass.name] = info
            disabled.update(_disabled)

    return result, disabled


def find_avocado_tests(path):
    return find_python_tests('avocado', 'Test', _determine_match_avocado, path)


def _determine_match_unittest(module, klass,
                              docstring):  # pylint: disable=W0613
    """
    Implements the match check for Python Unittests
    """
    return module.is_matching_klass(klass)


def find_python_unittests(path):
    found, _ = find_python_tests('unittest', 'TestCase',
                                 _determine_match_unittest,
                                 path)
    return found
